{****************************************************************************
 * GrafX2 PCI driver unit - copyright (c) 1999 by Stefan Goehler            *
 * Copyright details in grafx2.txt - without this file, this copy is        *
 * illegal!                                                                 *
 * References: Ralph Browns Interrupt List                                  *
 * Both PCI Bus types tested, should work without problems                  *
 ****************************************************************************}
unit drv_pci;

interface

uses ports;

type
  tpcidevinfo = record
    vendorid      : word;
    deviceid      : word;
    buscmd        : word;
    busstate      : word;
    revision      : byte;
    proginterface : byte;
    class         : word;
    cachelinesize : byte;
    latencytimer  : byte;
    headertype    : byte;
    unknown       : byte;
    base0         : longint;
    base1         : longint;
    base2         : longint;
    base3         : longint;
    base4         : longint;
    base5         : longint;
    reserved1     : array[0..7] of byte;
    rombase       : longint;
    reserved2     : array[0..7] of byte;
    intrline      : byte;
    intrpin       : byte;
    mingrant      : byte;
    maxlat        : byte;
    reserved3     : array[64..255] of byte;
  end;

function  pciavailable : boolean;
function  pcigetmaxdev : word;
function  pcifinddev(vendor,nu : word) : word;
function  pcigetdevinfo(dev : word;var info : tpcidevinfo) : boolean;

implementation

var
  l       : longint;
  pcitype : word; { PCI is available if this number is <> 0 }

function pciavailable : boolean;
begin
  pciavailable := pcitype <> 0;
end;

function pcigetmaxdev : word;
{ Returns maximum possible number of PCI devices. If no PCI
  bus found, it returns zero.                               }
begin
  case pcitype of
    1 : pcigetmaxdev := 512;
    2 : pcigetmaxdev := 15;
    else pcigetmaxdev := 0;
  end;
end;

function pcifinddev(vendor,nu : word) : word;
{ Preferable if you know what you wanna search. VENDOR stands
  for the Vendor ID of the card to search, NU stands for the
  number of the device with this ID. So it's possible to search
  also additional cards. The first value for NU is zero.
  Returns zero if no card with this ID was found (or NU is
  > 0 and there's no additional card anymore)                   }
var
  l,i : longint;
begin
  pcifinddev := 0;
  case pcitype of
    1 : for i := 0 to 511 do
        begin
          outpl($CF8,$80000000+i shl 11);
          l := inpl($CFC);
          if (l and $FFFF <> $FFFF)and(l shr 16 <> $FFFF)and
             (l and $FFFF = vendor) then if (nu = 0) then
          begin
            pcifinddev := i;
            break;
          end else dec(nu);
        end;
    2 : begin
          outpl($CF8,$80);
          outpl($CFA,0);
          for i := 0 to 15 do begin
            l := inpl(i*256+$C000);
            if (l and $FFFF <> $FFFF)and(l shr 16 <> $FFFF)and
               (l and $FFFF = vendor) then if (nu = 0) then
            begin
              pcifinddev := i;
              break;
          end else dec(nu);
          end;
          outpl($CF8,0);
        end;
  end;
end;

function pcigetdevinfo(dev : word;var info : tpcidevinfo) : boolean;
{ DEV is the device number you want information about. You can
  retrieve this number via findpcidev or you can do a generic
  search from 0 to pcigetmaxdev.                              }
begin
  pcigetdevinfo := false;
  case pcitype of
    1 : asm
          db 66h;xor   bx,bx
                 mov   bx,dev
          db 66h;shl   bx,11
          db 66h;or    bx,0000h;dw 8000h
                 mov   cx,64
                 les   di,info
                 @lp:
          db 66h;mov   ax,bx
                 mov   dx,0CF8h
          db 66h;out   dx,ax
                 mov   dx,0CFCh
          db 66h;in    ax,dx
          db 66h;mov   es:[di],ax
                 add   di,4
                 add   bx,4
                 dec   cx
                 jnz   @lp
        end;
    2 : asm
                 mov   dx,0CF8h
                 mov   al,$80
                 out   dx,al
                 mov   dx,0CFAh
                 xor   al,al
                 out   dx,al

                 mov   dx,dev
                 shl   dx,8
                 or    dx,0C000h
                 mov   cx,64
                 les   di,info
                 @lp:
          db 66h;in    ax,dx
          db 66h;mov   es:[di],ax
                 add   di,4
                 add   bx,4
                 dec   cx
                 jnz   @lp

                 mov   dx,0CF8h
                 xor   al,al
                 out   dx,al
        end;
    else exit;
  end;
  pcigetdevinfo := info.vendorid <> $FFFF;
end;

begin
  pcitype    := 0;
  port[$CF8] := 0;
  port[$CFA] := 0;
  if (port[$CF8] = 0)and(port[$CFA] = 0) then pcitype := 2
  else begin
    l := inpl($CF8);
    outpl($CF8,$80000000);
    if inpl($CF8) = $80000000 then pcitype := 1;
    outpl($CF8,l);
  end;
end.