{$A+,B-,F+,G+,N+,P+,Q-,R-,T-,V+,X+}
{****************************************************************************
 * VESA driver for GX2 - copyright (c) 1998-1999 by Stefan Goehler          *
 ****************************************************************************}
{$IFNDEF DPMI}
  !Real mode not supported!
  This unit is fully 32 bit oriented, so forget about real mode.
  You need BP7 to use protected mode. If you have it, please
  switch at compiler/platform to PM (start bp.exe!)
{$ENDIF}
unit drv_vesa;
interface
uses gx_glob,winapi,newfrontier;

implementation
type
  tvesainfo = record
    VESASignature : array[0..3] of char;
    VESAVersion   : word;
    OEMStringPtr  : pchar;
    Capabilities  : longint;
    VideoModePtr  : pointer;
    totalmemory   : word;
    { beware - this will only be filled if VBE 2+ is available }
    OemSoftwareRev    : word;
    OemVendorNamePtr  : pchar;
    OemProductNamePtr : pchar;
    OemProductRevPtr  : pchar;
    reserved          : array[1..222] of byte;
    OemData           : array[1..256] of byte;
  end;

  tmodeinfo = record
    ModeAttributes      : word;
    WinAAttributes      : byte;
    WinBAttributes      : byte;
    WinGranularity      : word;
    Winsize             : word;
    WinASegment         : word;
    WinBSegment         : word;
    WinFuncPtr          : pointer;
    BytesPerScanLine    : word;
    XResolution         : word;
    YResolution         : word;
    XCharsize           : byte;
    YCharsize           : byte;
    NumberOfPlanes      : byte;
    BitsPerPixel        : byte;
    NumberOfBanks       : byte;
    MemoryModel         : byte;
    BankSize            : byte;
    NumberOfImagePages  : byte;
    Reserved1           : byte;
    RedMaskSize         : byte;
    RedFieldPosition    : byte;
    GreenMaskSize       : byte;
    GreenFieldPosition  : byte;
    BlueMaskSize        : byte;
    BlueFieldPosition   : byte;
    RsvdMaskSize        : byte;
    RsvdFieldPosition   : byte;
    DirectColorModeInfo : byte;
    { VBE 2 Variables }
    PhysBasePtr         : longint;
    OffScreenMemOffset  : longint;
    OffScreenMemSize    : word;
    { VBE 3 Variables }
    LinBytesPerScanLine   : longint;
    BnkNumberOfImagePages : byte;
    LinNumberOfImagePages : byte;
    LinRedMaskSize        : byte;
    LinRedFieldPosition   : byte;
    LinGreenMaskSize      : byte;
    LinGreenFieldPosition : byte;
    LinBlueMaskSize       : byte;
    LinBlueFieldPosition  : byte;
    LinRsvdMaskSize       : byte;
    LinRsvdFieldPosition  : byte;
    MaxPixelClock         : longint;
    Reserved              : array[1..188] of byte;
  end;

  trealregs = record
    RealEDI  : longint;
    RealESI  : longint;
    RealEBP  : longint;
    Reserved : longint;
    RealEBX  : longint;
    RealEDX  : longint;
    RealECX  : longint;
    RealEAX  : longint;
    RealFlags: word;
    RealES   : word;
    RealDS   : word;
    RealFS   : word;
    RealGS   : word;
    RealIP   : word;
    RealCS   : word;
    RealSP   : word;
    RealSS   : word;
  end;

const
  vesa12 = $0102;
  vesa2  = $0200;
  vesa3  = $0300;

var
  vesainfo : tvesainfo;
  modeinfo : tmodeinfo;
  selector : longint;
  shifter  : byte;
  twowins  : boolean;
  modeptr  : pmode;

function intr32(intno : byte;var regs : trealregs) : boolean; assembler;
asm
  mov   ax,300h
  sub   bx,bx
  mov   bl,intno
  sub   cx,cx
  les   di,regs
  int   31h
  cmc
  mov   ax,0
  adc   ax,ax
end;

function getvesainfo(var vesainfo : tvesainfo) : boolean;
var
  regs : trealregs;
begin
  getvesainfo := false;
  selector := globaldosalloc(512);
  fillbyte(farptr(@regs),sizeof(trealregs),0);
  fillchar(regs,sizeof(regs),0);
  move2(farptr(@vesainfo),ptr48(selector,0),sizeof(vesainfo));
  with regs do begin
    reales  := selector shr 16;
    realedi := 0;
    realeax := $4F00;
  end;
  intr32($10,regs);
  move2(ptr48(selector,0),farptr(@vesainfo),sizeof(vesainfo));
  globaldosfree(selector);
  regs.realeax := regs.realeax and $FFFF;
  getvesainfo := regs.realeax = $4F;
end;


function getmodeinfo(var modeinfo : tmodeinfo;mode : word) : boolean;
var
  regs : trealregs;
begin
  selector := globaldosalloc(sizeof(tmodeinfo));
  fillbyte(ptr48(selector,0),sizeof(tmodeinfo),0);
  fillchar(regs,sizeof(regs),0);
  with regs do begin
    reales  := selector shr 16;
    realedi := 0;
    realeax := $4F01;
    realecx := mode;
  end;
  intr32($10,regs);
  move2(ptr48(selector,0),farptr(@modeinfo),sizeof(tmodeinfo));
  globaldosfree(selector);
  regs.realeax := regs.realeax and $FFFF;
  getmodeinfo := regs.realeax = $4F;
end;


procedure fillmodetable_vesa(var table);
type
  tvideomode = array[0..511] of word;
const
  supportedbpp : set of byte=[8,15,16,24,32];
var
  i,rtab    : word;
  videomode : ^tvideomode;
begin
  i          := 0;
  getmem(videomode,1024);
  move(ptr(getrmselector(longint(vesainfo.videomodeptr) shr 16),longint(vesainfo.videomodeptr) and $FFFF)^,
       videomode^,1024);
  for rtab := 0 to 4 do tmodetable(table)[rtab].avmodes := 0;
  while (videomode^[i] <> $FFFF)and(i < 512) do begin
    if getmodeinfo(modeinfo,videomode^[i]) then
    if (modeinfo.memorymodel <> 0)and(modeinfo.bitsperpixel in supportedbpp)
    then begin
      case modeinfo.bitsperpixel of
        8  : rtab := 0;
        15,16 : rtab := modeinfo.greenmasksize-4;
        24,32 : if modeinfo.rsvdmasksize = 8 then rtab := 4 else rtab := 3;
        else rtab := 5;
      end;
      if (rtab < 5)and(tmodetable(table)[rtab].avmodes < 25)and(
         longint(modeinfo.bytesperscanline)*modeinfo.yresolution <=
         longint(vesainfo.totalmemory) shl 16)and
         (getbit(modeinfo.modeattributes,4))and
         (getbit(modeinfo.modeattributes,0)) then begin
        tmodetable(table)[rtab].mode[tmodetable(table)[rtab].avmodes].resx := modeinfo.xresolution;
        tmodetable(table)[rtab].mode[tmodetable(table)[rtab].avmodes].resy := modeinfo.yresolution;
        tmodetable(table)[rtab].mode[tmodetable(table)[rtab].avmodes].mode := videomode^[i];
        inc(tmodetable(table)[rtab].avmodes);
      end;
    end;
    inc(i);
  end;
  freemem(videomode,1024);
end;

function initvesamode(mode : word;lfb : boolean) : boolean;assembler;
asm
           mov   bx,mode
           cmp   lfb,true
           jne   @nolfb   { Shall we use LFB?                         }
           or    bx,4000h { Yep - set LFB bit                         }
  @nolfb:  mov   ax,4F02h
           int   10h
           xor   ax,004Fh { ah <> 0 if mode fail, al <> 4F if no vesa }
           jnz   @fail
           mov   ax,1     { true                                      }
           jmp   @end
  @fail:   xor   ax,ax    { false                                     }
  @end:
end;

procedure setbank_vesa(bank : word);far;assembler;
{set the graphical bank of vesa}
asm
  push  ax
  push  bx
  push  cx
  push  dx
  mov   dx,bank
  cmp   dx,lastbank
  je    @end
  mov   lastbank,dx
  mov   cl,shifter

  shl   dx,cl
  xor   cl,cl
  xor   bx,bx
  mov   ax,4F05h
  int   10h
  cmp   twowins,0
  je    @end
    mov   bx,0001h
    mov   ax,4F05h
    int   10h
  @end:
  pop   dx
  pop   cx
  pop   bx
  pop   ax
end;

procedure decbank_vesa;far;assembler;
{decrease the graphical bank of vesa}
asm
  push  ax
  push  bx
  push  cx
  push  dx
  mov   dx,lastbank
  dec   dx
  mov   lastbank,dx
  mov   cl,shifter
  shl   dx,cl
  xor   cl,cl
  xor   bx,bx
  mov   ax,4F05h
  int   10h
  cmp   twowins,0
  je    @end
    mov   bx,0001h
    mov   ax,4F05h
    int   10h
  @end:
  pop   dx
  pop   cx
  pop   bx
  pop   ax
end;

procedure incbank_vesa;far;assembler;
{increase the graphical bank of vesa}
asm
  push  ax
  push  bx
  push  cx
  push  dx
  mov   dx,lastbank
  inc   dx
  mov   lastbank,dx
  mov   cl,shifter
  shl   dx,cl
  xor   cx,cx
  xor   bx,bx
  mov   ax,4F05h
  int   10h
  cmp   twowins,0
  je    @end
    mov   bx,0001h
    mov   ax,4F05h
    int   10h
  @end:
  pop   dx
  pop   cx
  pop   bx
  pop   ax
end;

procedure setdisplaypos_vesa(x,y : word);far;assembler;
asm
  mov   ax,4F07h
  xor   bx,bx
  mov   cx,x
  mov   dx,y
  int   10h
end;

function setvirtualscreenwidth_vesa(width : word) : word;far;
var
  state,bpl : word;
begin
  if (modeptr^.flags and mdlinear <> 0) then begin
    width := width*modeptr^.pixel;
    asm
      mov   ax,4F06h
      mov   bx,02h
      mov   cx,width
      int   10h
      mov   state,ax
      mov   bpl,bx
    end;
    setvirtualscreenwidth_vesa := bpl div modeptr^.pixel;
  end else
  begin
    asm
      mov   ax,4F06h
      xor   bx,bx
      mov   cx,width
      int   10h
      mov   state,ax
      mov   bpl,bx
    end;
    setvirtualscreenwidth_vesa := bpl;
  end;
  if (state <> $004F) then setvirtualscreenwidth_vesa := 0;
end;

function w2kactive : boolean;assembler;
asm
  mov   ax,160Ah
  int   2Fh
  xor   ax,ax
  cmp   bh,5
  jne   @end
  mov   ax,1
@end:
end;

function setupvideo_vesa(modenum : longint;var mode : tmode;flags : longint) : word;
var
  i : word;
begin
  modeptr := @mode;
  setupvideo_vesa := denoerror;
{  if w2kactive then flags := flags or mdbanked;}
  if (vesainfo.vesaversion >= vesa2)and
     ((flags = mdauto)or(flags and mdbanked = 0))and
     (getbit(modeinfo.modeattributes,7))and(modeinfo.physbaseptr <> 0) then
     mode.flags := mode.flags or mdlinear;
  mode.mode := modenum;
  if mode.mode = 0 then begin
    setupvideo_vesa := demdnotav;
    exit;
  end;
  getmodeinfo(modeinfo,mode.mode);
  mode.memsize   := longint(vesainfo.totalmemory) shl 16;

  asm mov ax,3; int 10h;end; { required by some S3 cards }
  if not initvesamode(mode.mode,mode.flags and mdlinear <> 0) then begin
    setupvideo_vesa := defail;
    exit;
  end;

  if ((mode.flags and mdlinear) = 0) then begin
    if (getbit(modeinfo.winAAttributes,1))and
       (getbit(modeinfo.winAAttributes,2)) then
       else begin
         twowins := true;
         mode.flags := mode.flags or mdtwowin;
       end;
    if getbit(modeinfo.winBAttributes,0) then begin
      if getbit(modeinfo.winBAttributes,2) then mode.writeptr := getrmselector(modeinfo.winBsegment);
      if getbit(modeinfo.winAAttributes,2) then mode.writeptr := getrmselector(modeinfo.winAsegment);
      if getbit(modeinfo.winBAttributes,1) then mode.readptr := getrmselector(modeinfo.winBsegment);
      if getbit(modeinfo.winAAttributes,1) then mode.readptr := getrmselector(modeinfo.winAsegment);
    end else
    begin
      mode.writeptr := getrmselector(modeinfo.winAsegment);
      mode.readptr  := mode.writeptr;
    end;
    lastbank  := 65535;
    shifter   := 7;
    i := modeinfo.wingranularity;
    repeat
      i := i shr 1;
      dec(shifter)
    until i = 0;
  end else begin
{    mode.writeptr := allocatedescriptor(1);
    setsegmentbaseaddress(mode.writeptr,PhysicalAddressMapping(modeinfo.physbaseptr,mode.memsize));
    setsegmentlimit(mode.writeptr,mode.memsize);
    mode.readptr   := mode.writeptr;}
    mode.linearptr := physicaladdressmapping(modeinfo.physbaseptr,mode.memsize);
    mode.writeptr  := createdescriptor(mode.linearptr,mode.memsize);
    mode.readptr   := mode.writeptr;
    i              := getsegmentaccessrights(mode.writeptr);
    setsegmentaccessrights(mode.writeptr,(i and $FF70) or $93);
  end;

  mode.bpp        := modeinfo.bitsperpixel;
  if modeinfo.greenmasksize = 5 then mode.bpp := 15;
  if (modeinfo.rsvdmasksize = 8)and(mode.bpp = 24) then mode.bpp := 32;

  mode.requiredlines := 0;
  mode.requiredmem   := 0;
  mode.pixel      := (mode.bpp+7) shr 3;
  mode.resx       := modeinfo.xresolution;
  mode.resy       := modeinfo.yresolution;
  mode.bpl        := modeinfo.bytesperscanline;
  mode.lasty      := mode.memsize div mode.bpl-1;
  mode.numofpages := mode.memsize div (longint(mode.bpl)*mode.resy);
  setupvideo_vesa := denoerror;
end;

procedure closevideo_vesa(var mode : tmode);
begin
  if (mode.flags and mdlinear <> 0)and(mode.writeptr <> 0) then begin
    freedescriptor(mode.writeptr);
    {freephysicaladdressmapping(mode.linearptr);}
  end;
  asm mov ax,textscreen.mode; int 10h;end;
end;

function detect_vesa : word;forward;

const
  vbedriver : tgfxdriver=(
  info:
   (drvname       : 'VBE 1.2+ Driver';
    drvversion    : $0036;
    drvsubversion : $0822;
    drvversionstr : '0.36.0822';
    drvprod       : 'Stefan Goehler/Crossfire';
    copyright     : '(c) 1998-2000 by Stefan Goehler';
    producer      : 'VESA';
    name          : '';
    rev           : 0;
    subrev        : 0;
    capabilities:
    (cinterface  : gdbanked+gdsetoffset+gdvwidth;
     cclipping   : 0;
     cbitmapping : 0;
     cprimitives : 0;
     cpolygons   : 0);
    memory        : 0;
    linearaddress : 0;
    bustype       : 0;
    priority      : 1);
  proc:
   (detect        : detect_vesa;
    fillmodetable : fillmodetable_vesa;
    fillrefreshtable : nil;
    setupvideo       : setupvideo_vesa;
    closevideo       : closevideo_vesa;
    setclip          : nil;
    setrefresh       : nil;
    fillrect         : nil;
    bitblit          : nil;
    scale            : nil;
    fillpoly         : nil;
    clear            : nil;
    setbank          : setbank_vesa;
    incbank          : incbank_vesa;
    decbank          : decbank_vesa;
    setdisplaypos    : setdisplaypos_vesa;
    setvirtualscreenwidth : setvirtualscreenwidth_vesa;
    savestate        : nil;
    restorestate     : nil;
    switchdacto8     : nil));

function detect_vesa : word;
begin
  detect_vesa := 0;
  vesainfo.vesasignature := 'VBE2';{ We want some VBE2+ info }
  if not getvesainfo(vesainfo) then exit;
  if (vesainfo.vesasignature = 'VESA')and(vesainfo.vesaversion >= vesa12) then
  { Signature OK and version >= 1.2? }
  begin { Oki, let's start... }
    detect_vesa := 1;
    with vbedriver.info do begin
      name     := inttostr(hi(vesainfo.vesaversion))+'.'+inttostr(lo(vesainfo.vesaversion));
      rev      := vesainfo.vesaversion;
      memory   := longint(vesainfo.totalmemory) shl 16;
      if (rev >= vesa2){and(not w2kactive)} then capabilities.cinterface :=
       capabilities.cinterface or gdlinear;
    end;
  end;
end;

begin
  registergfxdriver(@vbedriver);
end.

