{$A+,B-,F+,G+,N+,P+,Q-,R-,T-,V+,X+}
{****************************************************************************
 * GrafX2 utilities unit - copyright (c) 1999 by Stefan Goehler             *
 * Contains currently:                                                      *
 * -new page flipping support                                               *
 * -multiple rectangle putter (rects from an image to screen)               *
 ****************************************************************************}
unit gx_utils;
interface
uses grafx2,newfrontier,gx_glob;
type
  trect = record
    posx,posy   : longint;
    sizex,sizey : longint;
  end;

  pbuffer = ^tbuffer;
  tbuffer = object
    screen : pgrafx; { Screen where the application should draw to }
    constructor init(var scr : tgxscreen;buffers,flags : word);
    { Usual, easy flipping }
    procedure flip;  { Flip screen to drawn buffer }
    { Select buffer where to draw to: }
{    procedure selectbuffer(buffer : word);
    { Select buffer where to flip to: }
{    procedure flipto(buffer : word);}
    destructor done;
    private
    outscr : pgxscreen;             { Destination screen              }
    vimage : array[0..2] of timage; { VSCR-images                     }
    vscr   : array[0..2] of pgxmem; { Virtual screens                 }
    actpage   : integer;            { Actually used page              }
    actbuffer : integer;            { Actually used buffer            }
    numbuffer : integer;            { Number of used buffers          }
    scrbuf    : boolean;            { True if no virtual screens used }
    fl        : word;
  end;

const
  { BFxxxx-Flags for tbuffer.init }

  bfauto      = $0000;   { Let the program decide                         }
  bfmemonly   = $0001;   { Use only memory buffers                        }
  bfmem2scr   = $0002;   { Draw to memory, but use pageflipping           }
  bfretrace   = $0004;   { Memory-buffering: wait for retrace before draw }


procedure drawrects2screen(screen : pgrafx;scrposx,scrposy : longint;
                           image : timage;rect : array of trect;
                           numofrects : longint);

implementation

constructor tbuffer.init(var scr : tgxscreen;buffers,flags : word);
{ Init flags:

  bfauto    : uses screen pageflipping, if available pages >= buffers,
              otherwise normal memory-buffering (like bfmemonly)
  bfmemonly : use only memory pages for flipping
  bfmem2scr : uses 2 screen pages for flipping, draws into memory
              buffers, but draws buffers on invisible screen part and
              flips screen page thereafter
  bfretrace : only applies with bfmemonly, wait for retrace before drawing


  Buffers   : more than 2 make no sense if you don't use any interrupt-
              driven flipping (not supported by tbuffer-object)

  You need to draw to tbuffer.output, NOT to your normal screen!!
}
var
  i   : integer;
  err : word;
begin
  fl := flags;
  numbuffer := buffers;
  outscr := @scr;
  if (fl and (bfmemonly or bfmem2scr) = 0)and(scr.mode.numofpages >= buffers) then
  begin
    actpage := buffers-1;
    outscr^.setdrawpage(actpage);
    outscr^.setvisiblepage(0);
    screen := outscr;
    scrbuf := true;
  end else begin
    if (fl and bfmem2scr <> 0)and(scr.mode.numofpages >= 2) then
    begin
      actpage := 1;
      outscr^.setdrawpage(actpage);
      outscr^.setvisiblepage(0);
    end else fl := fl and not bfmem2scr;

    for i := 0 to numbuffer-2 do begin
      err := openimage(vimage[i],outscr^.maxx+1,outscr^.maxy+1,
                       outscr^.mode.bpp,1,imdefault);
      new(vscr[i],init(vimage[i]));
      if (err <> errnoerror)or(vscr[i] = nil) then
      begin
        for i := 0 to i do begin
          if vscr[i] <> nil then dispose(vscr[i],close);
          closeimage(vimage[i]);
        end;
        fail;
      end;
    end;
    actbuffer := buffers-2;
    screen := vscr[actbuffer];
    scrbuf := false;
  end;
end;

destructor tbuffer.done;
var
  i : integer;
begin
  if not scrbuf then for i := 0 to numbuffer-2 do
  begin
    dispose(vscr[i],close);
    closeimage(vimage[i]);
    if (fl and bfmem2scr <> 0) then
    begin
      outscr^.setdrawpage(0);
      outscr^.setvisiblepage(0);
    end;
  end else
  begin
    outscr^.setdrawpage(0);
    outscr^.setvisiblepage(0);
  end;
end;

procedure tbuffer.flip;
var
  oldwait : boolean;
begin
  if (scrbuf) then begin
    outscr^.setvisiblepage(actpage);
    actpage := (actpage+1) mod numbuffer;
    outscr^.setdrawpage(actpage);
  end else
  begin
    if (fl and bfretrace <> 0)and((fl and bfmem2scr = 0)) then waitretrace;

    if (outscr^.mode.flags and mdlinear <> 0) then
    begin
      movemul2screen(ptr48(vimage[actbuffer].data,vimage[actbuffer].imagestart),
                     ptr48(outscr^.mode.writeptr,outscr^.mode.displayofs),
                     vimage[actbuffer].bpl,outscr^.mode.bpl,
                     outscr^.mode.bpl,outscr^.maxy+1);
    end else outscr^.draw(0,0,vimage[actbuffer],drdefault);

    actbuffer := (actbuffer+1) mod (numbuffer-1);
    screen    := vscr[actbuffer];
    if (fl and bfmem2scr <> 0) then begin
      outscr^.setvisiblepage(actpage);
      actpage := (actpage+1) and 1;
      outscr^.setdrawpage(actpage);
    end;
  end;
end;

{procedure tbuffer.selectbuffer(buffer : word);
begin

end;

procedure tbuffer.flipto(buffer : word);
begin

end;}

procedure drawrects2screen(screen : pgrafx;scrposx,scrposy : longint;
                           image : timage;rect : array of trect;
                           numofrects : longint);
var
  i,y     : longint;
  scrptr,
  imgptr  : pointer48;
  scrpos  : word;
  imgpos  : longint;
  count   : longint;
  overlap : word;
begin
  if (screen^.mode.flags and mdlinear <> 0) then
  for i := 0 to numofrects-1 do begin
    scrptr := ptr48(screen^.mode.writeptr,screen^.mode.calcofs(scrposx+rect[i].posx,
                    scrposy+rect[i].posy,screen^.mode)+screen^.mode.displayofs);
    imgptr := ptr48(image.data,rect[i].posy*image.bpl+image.imagestart+
                    rect[i].posx*image.pixel);
    count := rect[i].sizex*image.pixel;
{    for y := 0 to rect[i].sizey-1 do
    begin
      move2screen(imgptr,scrptr,count);
      inc(tselofs32(scrptr).offset32,screen^.mode.bpl);
      inc(tselofs32(imgptr).offset32,image.bpl);
    end;}
    movemul2screen(imgptr,scrptr,count,image.bpl,screen^.mode.bpl,rect[i].sizey);
  end else
  for i := 0 to numofrects-1 do begin
    scrpos := screen^.mode.calcofs(scrposx+rect[i].posx,scrposy+rect[i].posy+
              screen^.mode.pageadd,screen^.mode);
    imgpos := rect[i].posy*image.bpl+image.imagestart+
              rect[i].posx*image.pixel;
    count := rect[i].sizex*image.pixel;
    for y := 0 to rect[i].sizey-1 do
    begin
      if scrpos < scrpos+word(count) then begin
        move2screen(ptr48(image.data,imgpos),ptr48(screen^.mode.writeptr,scrpos),count);
        asm
          les   di,screen
          mov   ax,es:[di].tmode.bpl.word
          add   scrpos,ax
          jnc   @next
            call   incbank
          @next:
        end;
      end else begin
        overlap := -scrpos;
        move2screen(ptr48(image.data,imgpos),ptr48(screen^.mode.writeptr,scrpos),overlap);
        incbank;
        move2screen(ptr48(image.data,imgpos+overlap),ptr48(screen^.mode.writeptr,0),count-overlap);
        inc(scrpos,screen^.mode.bpl);
      end;
      inc(imgpos,image.bpl);
    end;
  end;
end;

begin
end.