{$A+,B+,F-,G+,I-,O+,P+,Q-,R-,S-,T-,V-,X+,Y-}
{****************************************************************************
 * Author           : Stefan Goehler, Martin Beck, Oliver Fromme, Germany   *
 * Version          : 2.2                                                   *
 *******                                                                    *
 * my mail-adress   : stefan.goehler@gmx.de                                 *
 * my homepage      : http://sourcenet.home.pages.de                        *
 ****** COPYRIGHT ***********************************************************
 * This unit is delivered with the GX2-package. You are not allowed to      *
 * use this unit out of the GX2-package, i.e. for another graphical unit.   *
 * For more copyright information see gx2.txt                               *
 ****** HISTORY *************************************************************
 * Version 1.0  first version of this unit, delivered with GrafX 1.0        *
 * Version 1.1  *fixed a large amount of bugs                               *
 * Version 2.0  *adapted to GX2                                             *
 *              *loads now all non-interlaced images                        *
 *              *loads Photoshop PNGs                                       *
 *         2.1  +PNGLIB SAVES NOW PNG-FILES!! Big thanx to Martin Beck!     *
 *              *loader slightly faster                                     *
 *         2.2  *saver saves truecolor and grayscale images smaller         *
 *              *fixed some bugs                                            *
 ****************************************************************************}
unit pnglib;

interface

uses grafx2,gx_glob,gx_image,crc32,adler32,inflate,newfrontier,gx_col,
     zdeflate,zutil,zlib,stream32,objects;

  function  loadpng(var stream : tstream32;var image : timage;usebpp,flags : longint) : word;
  function  info_png(var stream : tstream32;var image_info : timageinfo) : word;
  function  savepng(var Stream : TStream32;var image : timage) : word;


implementation

  procedure resetvars;far;forward;
  function  MyHeapErrorFunc(Size: word) : integer;far;forward;
  procedure ResetReadBuffer;far;forward;
  function  CheckIO : word;far;forward;
  procedure ReadChunkHead;far;forward;
  function  FindChunk(c : string) : integer;far;forward;
  function  SkipChunk : word;far;forward;
  function  GetLineSize(PixelWidth : LongInt) : LongInt;far;forward;
  function  Process_MHDR : word;far;forward;
  function  Process_IHDR : word;far;forward;
  function  Process_PLTE : word;far;forward;
  function  Process_tRNS : word;far;forward;
  function  Process_IDAT : word;far;forward;
  function  Process_IEND : word;far;forward;
  function  Process_MEND : word;far;forward;

const
  idpng : array[0..7] of char = #137'PNG'#13#10#26#10;
  idmng : array[0..7] of char = #138'MNG'#13#10#26#10;

  numchunks = 16;
  chunks : array[1..numchunks] of
  record
    name    : array[0..3] of char;
    process : function  : word;
    haveit  : boolean; {True = chunk has appeared}
  end = ((name: 'MHDR'; process: Process_MHDR; HaveIt: false),
         (name: 'IHDR'; process: Process_IHDR; HaveIt: false),
         (name: 'PLTE'; process: Process_PLTE; HaveIt: false),
         (name: 'IDAT'; process: Process_IDAT; HaveIt: false),
         (name: 'IEND'; process: Process_IEND; HaveIt: false),
         (name: 'MEND'; process: Process_MEND; HaveIt: false),
         (name: 'gAMA'; process: skipchunk   ; HaveIt: false),
         (name: 'sBIT'; process: skipchunk   ; HaveIt: false),
         (name: 'cHRM'; process: skipchunk   ; HaveIt: false),
         (name: 'tRNS'; process: Process_tRNS; HaveIt: false),
         (name: 'bKGD'; process: skipchunk   ; HaveIt: false),
         (name: 'hIST'; process: skipchunk   ; HaveIt: false),
         (name: 'tEXt'; process: skipchunk   ; HaveIt: false),
         (name: 'zTXt'; process: skipchunk   ; HaveIt: false),
         (name: 'pHYs'; process: skipchunk   ; HaveIt: false),
         (name: 'tIME'; process: skipchunk   ; HaveIt: false));

  readbuffsize = 65528; { Reader constant }

  {color types}
  ctPalette = $01;
  ctColor   = $02;
  ctAlpha   = $04;

type
  treadbuff  = array[0..pred(readbuffsize)] of byte;
  preadbuff  = ^treadbuff;
  TChunkName = string[4];
var
  PNGStream       : pstream32;
  readbuff       : preadbuff;
  rbpos,rbend    : word;

  chunkhead : record
    length : longint;
    name   : array[0..3] of char
  end;

  pnghead : Record {see the PNG spec, draft #9}
    sizex,sizey   : longint;
    bitspersample : byte  ;
    colortype     : byte  ;
    cm,filter,il  : byte  ;
  end;

  mnghead : Record {Not fully specified, PSP5 uses e.g. only 12 bytes}
    sizex,sizey    : longint;
    tickspersecond : longint;
    nominallayer   : longint;
    nominalframe   : longint;
    nominalplay    : longint;
    simpprofile    : longint;
  end;

  ActualChunk : record
    Length : LongInt;
    CRC : TCRC;
  end;

  imagep   : preadbuff;
  linesize : LongInt;{number of bytes in one row incl. filter byte}

  runCRC : tCRC; {see unit CRC32 for details}
  runadl : tadler; {see unit Adler32 for details}

  currentoffset : longint;{byte position in current row}
  decompbytes   : longint;{counts decompressed bytes}
  bytesneeded   : longint;{bytes actually decompressed, should be = decompbytes}
  numlines      : longint;{number of scanlines = number of filter bytes}
  bytesperline  : word;
  currentpass   : byte;     {Adam7 pass, 0..6}
  PNGImage     : ^TImage;
  imgpos        : longint;
  convertdata   : tcnvproc;
  cnvtmp        : pcnvproc;
  row1,row2     : preadbuff;
  datarow,lastrow : preadbuff;

  {ImagePtr      : LongInt;}
  ScanLines     : Word;
  PNGPal        : Pointer;
  SaveBpp       : Byte;
  SaveLineSize  : LongInt;

function checkio : word;
var
  iores : integer;
begin
  checkio := errnoerror;
  iores   := ioresult;
  if iores <> 0 then
   if iores = 100 then
    checkio := errfalseend else checkio := errfileerror;
end;

procedure resetreadbuffer;
begin
  rbpos := readbuffsize;
  rbend := readbuffsize;
end;

procedure bufferedread(var desti;count : word);
var
  r     : word;
  readcount : longint;
label
  noload;
begin
  asm
    db 66h;xor   ax,ax
    db 66h;xor   bx,bx
    db 66h;xor   dx,dx
           mov  ax,rbpos
           mov  bx,count
           mov  cx,bx
    db 66h;add  bx,ax
           mov  dx,readbuffsize
    db 66h;cmp  bx,dx
           jna  noload
  end;
{      if rbpos+LongInt(c) > LongInt(ReadBuffSize) then begin}
  r := readbuffsize-rbpos;
  if rbpos < readbuffsize then
  move2(farptr(@readbuff^[rbpos]),farptr(readbuff),r);
  readcount := PNGStream^.read32(farptr(@readbuff^[r]),rbpos);
  if readcount < rbpos then rbend := r+readcount;
  rbpos := 0;
{      end;}
  asm
    mov ax,rbpos
    mov cx,count
    noload:
    mov dx,ds
    lds si,readbuff
    add si,ax
    les di,desti
    mov ax,cx
    shr cx,2
    jz @not4
    db 0F3h,66h,0A5h{rep movsd}
    @not4:
    mov cx,ax
    and cx,11b
    jz @end
    rep movsb
    @end:
    mov ds,dx
    mov bx,rbpos
    add bx,ax
    cmp bx,rbend
    jna @na
    mov inoutres,100
    jmp @a
    @na:
    add rbpos,ax
    @a:
  end;
end;

procedure bufferskipback(c : longint);
begin
  if rbpos >= c then dec(rbpos,c)
  else begin
    resetreadbuffer;
    PNGStream^.seek(PNGStream^.getpos-c);
  end;
end;

function readcheck(var desti; c : word) : word;
var
  err : word;
begin
  bufferedread(desti,c);
  err := checkio;
  dec(chunkhead.length,c);
  updatecrc32(runcrc,desti,c);
  readcheck := err;
end;

function myheaperrorfunc(size: word) : integer;
begin
  if size = 0 then
  myheaperrorfunc := 2 {success}
  else
  myheaperrorfunc := 1 {return NIL}
end {MyHeapErrorFunc};

procedure readchunkhead;
begin
  bufferedread(chunkhead,sizeof(chunkhead));
  checkio;
  with chunkhead do begin
    swapendian32(length);
    initcrc32(runcrc);
    updatecrc32(runcrc,name,4);
  end;
end;

{Skip to the end of the current chunk and check the CRC.}

function skipchunk : word;
var
  checkcrc : tcrc;
  b        : byte;
  err      : word;
begin
  with chunkhead do
  if length < 0 then begin
    bufferskipback(-length);
    err := checkio;
    skipchunk := err;
    if err <> errnoerror then exit;
  end else
  while chunkhead.length > 0 do begin
    err := readcheck(b,1);
    if err <> errnoerror then begin
      skipchunk := err;
      exit;
    end;
  end;
  bufferedread(checkcrc,4);
  skipchunk := checkio;
end;

{Callback for inflate:  feed an input byte to inflate.}

(*function png_readbyte : byte;far;
var
  b        : byte;
  err      : word;
begin
  while chunkhead.length = 0 do begin
    skipchunk;
    readchunkhead;
  end;
  readcheck(b,1);
  dec(chunkhead.length);
  updatecrc32(runcrc,b,1);
  png_readbyte := b;
end;*)
function png_readbyte : byte;far;
var
  b        : byte;
  err      : word;
begin
  while chunkhead.length = 0 do begin
    skipchunk;
    readchunkhead;
  end;
  dec(chunkhead.length);
  if rbpos < readbuffsize then begin
    png_readbyte := readbuff^[rbpos];
    inc(rbpos);
  end else begin
    bufferedread(b,1);
    err := ioresult;
    png_readbyte := b;
  end;
end;


{Apply a filter to a single row of pixels.}
var
  w    : word;
  bypp : word;
  bpp  : word;

procedure subfilt_decode;assembler;
asm
  les   di,datarow
  mov   cx,bytesperline
  dec   cx
  sub   cx,bypp
  inc   di
  mov   si,di
  add   di,bypp
  @lp:
    mov   al,es:[si]
    add   es:[di],al
    inc   di
    inc   si
    dec   cx
  jnz   @lp
end;

procedure SubFilt_Encode;assembler;
asm
  push  ds

  les   di,Row1
  inc   di
  mov   cx,bypp
  mov   dx,BytesPerLine
  sub   dx,cx
  lds   si,DataRow
  mov   bx,si
  rep   movsb
  or    dx,dx
  je    @end
  @lp:
    mov   al,ds:[si]
    sub   al,ds:[bx]
    mov   es:[di],al
    inc   di
    inc   si
    inc   bx
    dec   dx
  jnz   @lp

  @end:
  pop   ds
end;

procedure upfilt_decode;assembler;
asm
  les   di,datarow
  mov   cx,bytesperline
  dec   cx
  inc   di
  push  ds
  lds   si,lastrow
  inc   si
  @lp:
    mov   al,ds:[si]
    add   es:[di],al
    inc   di
    inc   si
    dec   cx
  jnz   @lp
  pop ds
end;

procedure UpFilt_Encode;assembler;
asm
  push  ds

  les   di,Row1
  inc   di
  mov   ax,word ptr LastRow+2
  db    8eh,0e8h              {mov gs,ax}
  mov   bx,word ptr LastRow
  mov   cx,BytesPerLine
  lds   si,DataRow
  @lp:
    mov   al,ds:[si]
    db    65h,2ah,07h         {sub   al,gs:[bx]}
    mov   es:[di],al
    inc   si
    inc   di
    inc   bx
    dec   cx
  jnz   @lp

  pop ds
end;

procedure averagefilt_decode;assembler;
asm
  mov   dx,bytesperline
  mov   cx,bypp
  sub   dx,cx
  jz    @end
  les   di,datarow
  inc   di
  mov   bx,di
  push  ds
  lds   si,lastrow
  inc   si
  @lp:
    mov   al,ds:[si]
    shr   al,1
    add   es:[di],al
    inc   si
    inc   di
  dec   cx
  jnz   @lp

  xor   ax,ax
  cmp   dx,1
  jle   @stop
  dec   dx
  @lp2:
    mov   al,es:[bx]
    mov   cl,ds:[si]
    add   ax,cx
    shr   ax,1
    add   es:[di],al
    inc   si
    inc   di
    inc   bx
  dec   dx
  jnz   @lp2
  @stop:
  pop   ds
  @end:
end;

procedure AverageFilt_Encode;assembler;
asm
  push  ds

  les   di,Row1
  inc   di
  db    0fh,0b5h,1eh;dw LastRow   {lgs bx,LastRow}
  mov   cx,bypp
  mov   dx,BytesPerLine
  sub   dx,cx
  lds   si,DataRow
  push  bp
  mov   bp,si
  @lp:
    mov   al,ds:[si]
    db    65h,8ah,27h             {mov ah,gs:[bx]}
    shr   ah,1
    sub   al,ah
    mov   es:[di],al
    inc   si
    inc   di
    inc   bx
    dec   cx
  jnz   @lp
  or    dx,dx
  je    @end
  @lp2:
    xor   ah,ah
    xor   ch,ch
    mov   cl,ds:[bp]
    db    65h,8ah,07h             {mov al,gs:[bx]}
    add   ax,cx
    shr   ax,1
    neg   al
    add   al,ds:[si]
    mov   es:[di],al
    inc   si
    inc   di
    inc   bx
    inc   bp
    dec   dx
  jnz @lp2

  @end:
  pop   bp
  pop   ds
end;

function paethpredictor(a,b,c : word) : byte;assembler;
var
  p,pa,pb,pc : word;
asm
  mov   si,b
  mov   di,a
  push  bp
  mov   bp,c
  mov   dx,di
  add   di,si
  sub   di,bp

  mov   ax,di
  sub   ax,dx
  jns   @nexta;neg ax;@nexta:
  mov   bx,di
  sub   bx,si
  jns   @nextb;neg bx;@nextb:
  mov   cx,di
  sub   cx,bp
  jns   @nextc;neg cx;@nextc:

  cmp   ax,bx
  jnbe  @useb
  cmp   ax,cx
  jnbe  @useb
  mov   ax,dx
  jmp   @end
  @useb:
  cmp   bx,cx
  jnbe  @usec
  mov   ax,si
  jmp   @end
  @usec:
  mov   ax,bp
  @end:
  pop   bp
end;

procedure paethfilt_decode;
var
  c1,c2 : byte;
begin
  c1 := 0;
  c2 := 0;
  for w := 0 to bytesperline-2 do begin
    if w >= bypp then begin
      c1 := datarow^[w-bypp+1];
      c2 := lastrow^[w-bypp+1];
    end;
    inc(datarow^[w+1],
    paethpredictor(c1,lastrow^[w+1],c2));
  end;
end;

procedure PaethFilt_Encode;
var
  c1,c2 : Byte;
begin
  c1:=0;
  c2:=0;
  for w:=1 to BytesPerLine do
  begin
    if w>bypp then
    begin
      c1:=DataRow^[w-bypp-1];
      c2:=LastRow^[w-bypp-1];
    end;
    Row1^[w]:=DataRow^[w-1]-PaethPredictor(c1,LastRow^[w-1],c2);
  end;
end;

procedure applyfilter;
begin
  case datarow^[0] of
    0 :;
    1 : subfilt_decode;
    2 : upfilt_decode;
    3 : averagefilt_decode;
    4 : paethfilt_decode;
  end;
end;

procedure ApplyEncodeFilter;
begin
  case Row1^[0] of
    0 : Move2(FarPtr(DataRow),FarPtr(@Row1^[1]),BytesPerLine);
    1 : SubFilt_Encode;
    2 : UpFilt_Encode;
    3 : AverageFilt_Encode;
    4 : PaethFilt_Encode;
  end;
end;

function getlinesize(pixelwidth : longint) : longint;
begin
  with pnghead do
  case colortype of
    0 : getlinesize := (bitspersample*pixelwidth+7) shr 3+1;
    2 : getlinesize := (bitspersample shr 3)*3*pixelwidth+1;
    3 : getlinesize := (bitspersample*pixelwidth+7) shr 3+1;
    4 : getlinesize := (bitspersample shr 2)*pixelwidth  +1;
    6 : getlinesize := (bitspersample shr 1)*pixelwidth  +1;
    else
    getlinesize := pixelwidth shl 3+1;
    {should be pretty save if user wants to ignore error}
  end;
end;

{Tables for Adam7 interlacing.}

const
  adam7_startrow : array[0..6] of byte=(0,0,4,0,2,0,1);
  adam7_startcol : array[0..6] of byte=(0,4,0,2,0,1,0);
  adam7_incrmrow : array[0..6] of byte=(8,8,8,4,4,2,2);
  adam7_incrmcol : array[0..6] of byte=(8,8,4,4,2,2,1);

var
 cury : longint;

{Number of pixels/row in current pass}

function passwidth : longint;
begin
  passwidth :=
  (pnghead.sizex+adam7_incrmcol[currentpass]-1-adam7_startcol[currentpass])
  div adam7_incrmcol[currentpass];
end;

{Number of rows in current pass}

function passheight : longint;
begin
  passheight :=
  (pnghead.sizey+adam7_incrmrow[currentpass]-1-adam7_startrow[currentpass])
  div adam7_incrmrow[currentpass]
end;

procedure showline(x,y : word;width : integer;bw : boolean;data : pointer);
{!! doesn't recognize bpp > 32 }
var
  p : pointer;
begin
  cnvypos := y;
  cnvxpos := x;
  imgpos  := PNGImage^.bpl*longint(y)+PNGImage^.imagestart+x*PNGImage^.pixel;
  convertdata(farptr(data),ptr48(PNGImage^.data,imgpos),width*bypp);
{  if bw then
  case bpp of
    1,2,4,8 : image_putline(x,y,width,bpp,data);
    16,32 : begin
           getmem(p,width);
           asm
             mov   cx,width
             mov   bx,bpp
             shr   bx,3
             les   di,data
             mov   dx,ds
             lds   si,p
             @lp:
               mov   al,es:[di]
               mov   ds:[si],al
               add   di,bx
               inc   si
             dec   cx
             jnz   @lp
             mov   ds,dx
           end;
           image_putline(x,y,width,8,p);
           freemem(p,width);
         end;
  end else
  case bpp of
    1,2,4,8,24 : image_putline(x,y,width,bpp,data);
    48,64 : begin
           getmem(p,pnghead.width*3);
           asm
             mov   cx,width
             mov   bx,bpp
             shr   bx,3
             les   di,data
             mov   dx,ds
             lds   si,p
             @lp:
               mov   al,es:[di]
               mov   ah,es:[di+2]
               mov   ds:[si],ax
               mov   al,es:[di+4]
               mov   ds:[si+2],al
               add   di,bx
               add   si,3
             dec   cx
             jnz   @lp
             mov   ds,dx
           end;
           image_putline(x,y,width,24,p);
           freemem(p,pnghead.width*3);
         end;
    32 : begin
           getmem(p,pnghead.width*3+1);
           asm
             mov   cx,width
             les   di,data
             mov   dx,ds
             lds   si,p
             @lp:
               db    66h;mov   ax,es:[di]
               db    66h;mov   ds:[si],ax
               add   di,4
               add   si,3
             dec   cx
             jnz   @lp
             mov   ds,dx
           end;
           image_putline(x,y,width,24,p);
           freemem(p,pnghead.width*3+1);
         end;
  end;}
end;

function png_flush(w : word) : integer;far;
var
  copyoffset,copycount,x,xp : word;
  datap                     : pointer;
begin
  png_flush  := 0;
  copyoffset := 0;
  inc(decompbytes,w);
  updateadler32(runadl,slide^,w);
  if currentpass > 6 then exit; {Process_IDAT detects this}
  while w > 0 do begin
    if pnghead.il = 1 then begin {interlaced}
      {Skip empty passes}
      while ((passwidth = 0)or(passheight = 0))and(currentpass < 7) do
      inc(currentpass);
      if currentpass > 6 then exit;
      bytesperline := getlinesize(passwidth);
    end else {non-interlaced}
    bytesperline := linesize;
    if w > bytesperline-currentoffset then
    copycount := bytesperline-currentoffset else copycount := w;
    if currentoffset+copycount <= 65528 then

    move2(farptr(@slide^[copyoffset]),farptr(@datarow^[currentoffset]),copycount);
    dec(w,copycount);
    inc(copyoffset,copycount);
    inc(currentoffset,copycount);
    if currentoffset >= bytesperline then begin {next row}
      applyfilter;
      currentoffset := 0;
      if pnghead.il = 1 then begin {interlaced}
        if currentpass < 6 then begin
          xp    := adam7_startcol[currentpass];
          datap := ptr(seg(datarow^),ofs(datarow^)+1);
          for x := 0 to passwidth do begin
            showline(xp,cury,1,pnghead.colortype in [0,4],datap);
            inc(xp,adam7_incrmcol[currentpass]);
            inc(word(datap),bypp);
          end;
        end else
        showline(0,cury,pnghead.sizex,pnghead.colortype in [0,4],
                 ptr(seg(datarow^),ofs(datarow^)+1));
        inc(cury,adam7_incrmrow[currentpass]);
        if cury >= pnghead.sizey then begin
          inc(currentpass);
          if currentpass > 6 then exit;
          cury := adam7_startrow[currentpass];
        end;
        if pnghead.colortype <> 3 then swaplong(datarow,lastrow);
         {move2(ptr48(seg(datarow^),ofs(datarow^)+1),farptr(@lastrow^),linesize-1);}
      end else begin {non-interlaced}
        showline(0,cury,pnghead.sizex,pnghead.colortype in [0,4],
                 ptr(seg(datarow^),ofs(datarow^)+1));
        if pnghead.colortype <> 3 then swaplong(datarow,lastrow);
         {move2(ptr48(seg(datarow^),ofs(datarow^)+1),farptr(@lastrow^),linesize-1);}
        inc(cury);
        if cury >= pnghead.sizey then exit;
      end;
    end;
  end;
end;

function findchunk(c : string) : integer;
var
   i : integer;
begin
  findchunk := -1;
  for i := 1 to numchunks do
  if c = chunks[i].name then begin
    findchunk := i;
    break;
  end;
end;

function checklength(l : longint) : longint; {returns actual length}
begin
  checklength := chunkhead.length;
end;


function process_mhdr : word;
var
  w,err : word;
begin
  checklength(sizeof(pnghead));
  fillbyte(farptr(@pnghead),sizeof(pnghead),0);
  err := readcheck(pnghead,sizeof(pnghead));
  process_mhdr := err;
  if err <> errnoerror then exit;
  with pnghead do begin
    swapendian32(sizex);
    swapendian32(sizey);
    linesize := getlinesize(sizex);
  end;
  if linesize > 65528 then
    w := 65528
  else
  w := linesize;
  getmem(row1,w);
  getmem(row2,w);
  if (row1 = nil)or(row2 = nil) then begin
    process_mhdr := errnomem;
    exit;
  end;
  fillbyte(farptr(row1),w,0);
  fillbyte(farptr(row2),w,0);
  process_mhdr := skipchunk;
end;

function process_ihdr : word;
var
  err : word;
begin
  checklength(sizeof(pnghead));
  fillbyte(farptr(@pnghead),sizeof(pnghead),0);
  err := readcheck(pnghead,sizeof(pnghead));
  process_ihdr := err;
  if err <> errnoerror then exit;
  with pnghead do begin
    swapendian32(sizex);
    swapendian32(sizey);
    linesize := getlinesize(sizex);
  end;
  process_ihdr := skipchunk;
end;


function process_trns : word;
var
  err,i : word;
  index : array[0..255] of byte;
begin
  PNGImage^.transcol := 0;
  if chunkhead.length <= 255 then begin
    err := readcheck(index,chunkhead.length);
    process_trns := err;
    if err <> errnoerror then exit;
    case pnghead.colortype of
      0 : PNGImage^.transcol := index[0]+word(index[1]) shl 8;
      2 : PNGImage^.transcol := index[0]+word(index[2]) shl 8+longint(index[4]) shl 16;
      3 : begin
            for i := 0 to chunkhead.length do if index[i] < 255 then break;
            PNGImage^.transcol := i;
          end;
    end;
    convertdata(farptr(@PNGImage^.transcol),farptr(@PNGImage^.transcol),
                chunkhead.length);
    PNGImage^.flags := PNGImage^.flags or imsprite;
  end;
  process_trns := skipchunk;
end;

function process_plte : word;
type
  tpal = array[0..767] of byte;
var
  dummy : longint;
  i     : byte;
  pal   :^tpal;
  palettesize : word;
  err   : word;
begin
  new(pal);
  with chunkhead do palettesize := length div 3;
  dummy := chunkhead.Length;
  process_plte := skipchunk;
  bufferskipback(dummy+4);
  bufferedread(pal^,dummy);
  pal_24to32(farptr(pal),cnvpal,palettesize);
  case PNGImage^.bpp of
    15 : pal_32to15(cnvpal,cnvpal,palettesize);
    16 : pal_32to16(cnvpal,cnvpal,palettesize);
  end;
  bufferedread(dummy,4);
  dispose(pal);
  process_plte := errnoerror;
{  if (pstart = 0)and(pend = 255) then setpal
  else for i := pstart to pend do setrgbpalette(i,pal^[i].r,pal^[i].g,pal^[i].b);}
end;

function process_idat : word;
var
  adlercheck : longint;
  result     : integer;
  w          : word   ;
begin
  process_idat := errnoerror;
  if chunks[findchunk(chunkhead.name)].haveit then begin
    skipchunk;
    exit;
  end;
  w := png_readbyte;
  w := (w shl 8) or png_readbyte;
  initadler32(runadl);
  decompbytes := 0;
  with pnghead do
  if il = 1 then begin {interlaced}
    bytesneeded := 0;
    numlines    := 0;
    for currentpass := 0 to 6 do begin
      inc(bytesneeded,getlinesize(passwidth)*passheight);
      inc(numlines,passheight)
    end
  end
  else begin {non-interlaced}
    bytesneeded := linesize*sizey;
    numlines    := sizey;
  end;
  currentoffset := 0;
  currentpass   := 0;
  cury          := 0;
  inflateread  := png_readbyte;
  inflateflush := png_flush;
  result := inflaterun;
{  for w := 1 to 4 do adlercheck := (adlercheck shl 8) or png_readbyte;}
  process_idat := skipchunk;
end;

function process_iend : word;
begin
  checklength(0);
  process_iend := skipchunk;
end;

function process_mend : word;
begin
  checklength(0);
  process_mend := skipchunk;
end;

procedure resetvars;
var
  i : byte;
begin
  linesize      := 0;
  currentoffset := 0;
  decompbytes   := 0;
  bytesneeded   := 0;
  numlines      := 0;
  currentpass   := 0;
  for i := 1 to numchunks do chunks[i].haveit := false;
  cury          := 0;
  chunkhead.name[0] := #0;
  lastrow   := nil;
  slide     := nil;
  readbuff  := nil;
  datarow := nil;
end;

function getbpp(coltype,bitdepth : byte) : byte;
begin
  case coltype of
    0,3 : getbpp := bitdepth;     { Grayscale/Palette       }
    2 : getbpp := bitdepth*3;     { RGB                     }
    4 : getbpp := bitdepth shl 1; { Greyscale, Alphachannel }
    6 : getbpp := bitdepth shl 2; { RGB, Alphachannel       }
  end;
end;

function bppvalid(coltype,bitdepth : byte) : boolean;
begin
  bppvalid := false;
  if bitdepth > 8 then exit; {!!}
  case coltype of
    0 : ; { Grayscale/Palette       }
    2 : ; { RGB                     }
    3 : ; { Palette image           }
    4 : ; { Greyscale, Alphachannel }
    6 : ; { RGB, Alphachannel       }
    else exit;
  end;
  bppvalid := true;
end;

function info_png(var stream : tstream32;var image_info : timageinfo) : word;
var
  BufMag    : array[0..7] of char;
  i         : integer;
  oldheaperrorfunc : pointer;

procedure prepareexit(code : word);
begin
  if readbuff <> nil then freemem(readbuff,readbuffsize);
  info_png := code;
  heaperror := oldheaperrorfunc;
  PNGStream^.done;
end;

begin
  info_png := errfileerror;
  if (stream.status <> stok) then exit;
  PNGStream := @stream;
  resetvars;
  oldheaperrorfunc := HeapError;
  HeapError := @MyHeapErrorFunc;
  getmem(ReadBuff,ReadBuffSize);
  if ReadBuff = NIL then begin;prepareexit(errnomem);exit;end;
  ResetReadBuffer;
  BufferedRead(BufMag,8);
  CheckIO;
  if BufMag <> idpng then begin;prepareexit(errfalseformat);exit;end;
  ReadChunkHead;
  if ChunkHead.Name = 'IHDR' then begin
    Process_IHDR;
    with image_info do begin
      sizex        := pnghead.sizex;
      sizey        := pnghead.sizey;
      bpp          := getbpp(pnghead.colortype,pnghead.bitspersample);
      greyscale    := pnghead.colortype in [0,4];
      alphachannel := pnghead.colortype in [4,6];
      transparency := false;
      compressed   := true;
      comptype     := 'LZ77';
      interlaced   := pnghead.il = 1;
      frames       := 1;
      trans_color  := 0;
      memsize      := longint(sizex)*sizey*bpp div 8;{!!}
      disksize     := PNGStream^.getsize;
      supported    := bppvalid(pnghead.colortype,pnghead.bitspersample);
      version      := 1;
    end;
  end else begin
    prepareexit(errfalseformat);
    exit;
  end;
  prepareexit(errnoerror);
end;

procedure setrgb(pal : pointer48;i,r,g,b : byte);assembler;
asm
  db 66h;les   di,pal.dword
  db 66h;xor   ax,ax
         mov   al,i
         shl   ax,2
  db 66h;add   di,ax
  db 66h;xor   ax,ax
         mov   al,r
  db 66h;shl   ax,16
         mov   ah,g
         mov   al,b
  db 26h,67h,66h,89h,07h {mov   es:[edi],eax}
end;

procedure setgreypal;
var
  i   : integer;
  bpp : word;
begin
  bpp := pnghead.bitspersample;
  if bpp >= 16 then bpp := 8;
  if pnghead.bitspersample = 1 then
  setrgb(cnvpal,1,255,255,255) else
  for i := 0 to (1 shl bpp)-1 do
  setrgb(cnvpal,i,i shl (8-bpp),i shl (8-bpp),i shl (8-bpp));
  case PNGImage^.bpp of
    15 : pal_32to15(cnvpal,cnvpal,word(1) shl bpp);
    16 : pal_32to16(cnvpal,cnvpal,word(1) shl bpp);
  end;
end;

function loadpng(var stream : tstream32;var image : timage;usebpp,flags : longint) : word;
var
  bufmag    : array[0..7] of char;
  i         : integer;
  oldheaperrorfunc : pointer;
  err,w     : word;

procedure prepareexit(code : word);
begin
  if row1     <> nil then freemem(row1,linesize);
  if row2     <> nil then freemem(row2,linesize);
  if slide    <> nil then freemem(slide,wsize);
  if readbuff <> nil then freemem(ReadBuff,ReadBuffSize);
  heaperror := oldheaperrorfunc;
  stream.done;
  loadpng  := code;
end;

begin
  loadpng := errfileerror;
  if (stream.status <> stok) then exit;
  PNGStream := @stream;
  resetvars;
  oldheaperrorfunc := heaperror;
  heaperror := @myheaperrorfunc;
  getmem(slide,wsize);
  getmem(readbuff,readbuffsize);
  if (slide = nil)or(readbuff = nil) then begin;prepareexit(errnomem);exit;end;
  resetreadbuffer;
  bufferedread(bufmag,8);
  checkio;
  if bufmag <> idpng then begin;prepareexit(errfalseformat);exit;end;

  readchunkhead;
  if chunkhead.name = 'IHDR' then begin
    err := process_ihdr;
    if err <> errnoerror then begin;prepareexit(err);exit;end;
    if not bppvalid(pnghead.colortype,pnghead.bitspersample) then
    begin;prepareexit(errfalseformat);exit;end;

    if linesize > 65528 then
      w := 65528
    else
    w := linesize;
    getmem(row1,w);
    getmem(row2,w);
    if (row1 = nil)or(row2 = nil) then begin
      prepareexit(errnomem);
      exit;
    end;
    fillbyte(farptr(row1),w,0);
    fillbyte(farptr(row2),w,0);

    datarow := row1;
    lastrow := row2;
    bpp  := getbpp(pnghead.colortype,pnghead.bitspersample);
    bypp := linesize div pnghead.sizex;
    if pnghead.colortype in [4,6] then flags := flags or ldhasalpha;
    cnvtmp := prepareimage_cnv(image,pnghead.sizex,pnghead.sizey,bpp,usebpp,0,flags,cnvimgmd+cnvswap);
    if cnvtmp = nil then begin;prepareexit(errfalsebpp);exit;end;
    convertdata := tcnvproc(cnvtmp);
    PNGImage := @image;
    if pnghead.colortype in [0,4] then begin
      pngimage^.flags := pngimage^.flags or imgreyscale;
      setgreypal;
    end;
    if (pnghead.colortype in [4,6])and(image.bpp = 32) then image.flags := image.flags or imalpha;
  end
  else begin
    prepareexit(errfalseformat);
    exit;
  end;

  with chunkhead do
  while name <> 'IEND' do begin
   readchunkhead;
   i := findchunk(chunkhead.name);
   if i >= 0 then begin
     err := chunks[i].process;
     chunks[i].haveit := true;
     if err <> errnoerror then begin;prepareexit(err);exit;end;
   end
   else begin
     err := skipchunk;
     if err <> errnoerror then begin;prepareexit(err);exit;end;
   end;
  end;
  prepareexit(errnoerror);
end;

function loadmng(var stream : tstream32;var image : timage;usebpp,flags : longint) : word;
var
  bufmag    : array[0..7] of char;
  i         : integer;
  oldheaperrorfunc : pointer;
  err,w     : word;

procedure prepareexit(code : word);
begin
  if row1     <> nil then freemem(row1,linesize);
  if row2     <> nil then freemem(row2,linesize);
  if slide    <> nil then freemem(slide,wsize);
  if readbuff <> nil then freemem(ReadBuff,ReadBuffSize);
  heaperror := oldheaperrorfunc;
  stream.done;
  loadmng  := code;
end;

begin
  PNGStream := @stream;
  resetvars;
  oldheaperrorfunc := heaperror;
  heaperror := @myheaperrorfunc;
  getmem(slide,wsize);
  getmem(readbuff,readbuffsize);
  if (slide = nil)or(readbuff = nil) then begin;prepareexit(errnomem);exit;end;

  resetreadbuffer;
  bufferedread(bufmag,8);
  checkio;
  if bufmag <> idpng then begin;prepareexit(errfalseformat);exit;end;

  readchunkhead;
  if chunkhead.name = 'IHDR' then begin
    err := process_ihdr;
    if err <> errnoerror then begin;prepareexit(err);exit;end;
    if not bppvalid(pnghead.colortype,pnghead.bitspersample) then
    begin;prepareexit(errfalseformat);exit;end;

    if linesize > 65528 then
      w := 65528
    else
    w := linesize;
    getmem(row1,w);
    getmem(row2,w);
    if (row1 = nil)or(row2 = nil) then begin
      prepareexit(errnomem);
      exit;
    end;
    fillbyte(farptr(row1),w,0);
    fillbyte(farptr(row2),w,0);

    datarow := row1;
    lastrow := row2;
    if pnghead.colortype in [0,4] then setgreypal;
    bpp  := getbpp(pnghead.colortype,pnghead.bitspersample);
    bypp := linesize div pnghead.sizex;
    if pnghead.colortype in [4,6] then flags := flags or ldhasalpha;
    cnvtmp := prepareimage_cnv(image,pnghead.sizex,pnghead.sizey,bpp,usebpp,0,flags,cnvimgmd+cnvswap);
    if cnvtmp = nil then begin;prepareexit(errfalsebpp);exit;end;
    convertdata := tcnvproc(cnvtmp);
    PNGImage := @image;
    if (pnghead.colortype in [4,6])and(image.bpp = 32) then image.flags := image.flags or imalpha;
  end
  else begin
    prepareexit(errfalseformat);
    exit;
  end;

  with chunkhead do
  while name <> 'MEND' do begin
   readchunkhead;
   i := findchunk(chunkhead.name);
   if i >= 0 then begin
     err := chunks[i].process;
     chunks[i].haveit := true;
     if err <> errnoerror then begin;prepareexit(err);exit;end;
   end
   else begin
     err := skipchunk;
     if err <> errnoerror then begin;prepareexit(err);exit;end;
   end;
  end;
  prepareexit(errnoerror);
end;

function StreamToGX2Err(var s : TStream32) : Word;
begin
  case s.Status of
    stOk : StreamToGX2Err:=errNoError;
    stError : StreamToGX2Err:=errFileError;
    stReadError : StreamToGX2Err:=errFileError;
    stWriteError : if s.ErrorInfo=101 then
                     StreamToGX2Err:=errNoDiskMem
                   else
                     StreamToGX2Err:=errWriteError;
{    stEndOfFile : StreamToGX2Err:=errFalseEnd;}
  end;
end;

function StartChunk(Name : TChunkName) : Word;
begin
  StartChunk:=ErrNoError;
  ActualChunk.Length:=0;
  InitCRC32(ActualChunk.CRC);

  PNGStream^.Write32(FarPtr(@Name),4);             {later filled with Length}
  if PNGStream^.Status<>stOk then
  begin
    StartChunk:=StreamToGX2Err(PNGStream^);
    Exit;
  end;

  PNGStream^.Write32(FarPtr(@Name[1]),4);          {save chunk name}
  if PNGStream^.Status<>stOk then
  begin
    StartChunk:=StreamToGX2Err(PNGStream^);
    Exit;
  end;
  UpdateCRC32(ActualChunk.CRC,Name[1],4);
end;

function WriteChunk(var Source;Length : Word) : Word;
begin
  WriteChunk:=ErrNoError;
  if Length=0 then
    Exit;
  UpdateCRC32(ActualChunk.CRC,Source,Length);
  Inc(ActualChunk.Length,Length);
  PNGStream^.Write32(FarPtr(@Source),Length);
  if PNGStream^.Status<>stOk then
  begin
    WriteChunk:=StreamToGX2Err(PNGStream^);
    Exit;
  end;
end;

function EndChunk : Word;
var
  Len : LongInt;
begin
  EndChunk:=ErrNoError;
  ActualChunk.CRC:=FinalCRC32(ActualChunk.CRC);
  PNGStream^.Seek(PNGStream^.GetPos-ActualChunk.Length-8);
  Len:=ActualChunk.Length;
  SwapEndian32(Len);
  PNGStream^.Write32(FarPtr(@Len),4);
  if PNGStream^.Status<>stOk then
  begin
    EndChunk:=StreamToGX2Err(PNGStream^);
    Exit;
  end;
  PNGStream^.Seek(PNGStream^.GetPos+ActualChunk.Length+4);
  PNGStream^.Reset;
  Swapendian32(ActualChunk.CRC);
  PNGStream^.Write32(FarPtr(@ActualChunk.CRC),4);
  if PNGStream^.Status<>stOk then
  begin
    EndChunk:=StreamToGX2Err(PNGStream^);
    Exit;
  end;
end;

function SavePalette : Word;

procedure PrepareExit(Code : Word);
begin
  if PNGPal<>nil then
    FreeMem(PNGPal,3*(1 shl SaveBpp));
  SavePalette:=Code;
end;

var
  Err : Word;
begin
  SavePalette:=ErrNoError;
  GetMem(PNGPal,3*(1 shl SaveBpp));
  if PNGPal=nil then begin SavePalette:=ErrNoMem;Exit;end;
  Pal_32to24_swap(Ptr48(PNGImage^.Data,PNGImage^.PalStart),farptr(PNGPal),1 shl SaveBpp);
  Err:=StartChunk('PLTE');
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
  Err:=WriteChunk(PNGPal^,3*(1 shl SaveBpp));
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
  Err:=EndChunk;
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
  PrepareExit(errNoError);
end;

function SaveTransparency : Word;
var
  TransCol : LongInt;
  Data : Byte;
  i : Word;
  Err : Word;
begin
  SaveTransparency:=ErrNoError;
  Err:=StartChunk('tRNS');
  if Err<>ErrNoError then begin SaveTransparency:=Err;Exit;end;
  TransCol:=PNGImage^.TransCol;
  if SaveBpp<=8 then
  begin
    Data:=255;
    if Transcol<>0 then
      for i:=1 to TransCol do
      begin
        Err:=WriteChunk(Data,1);
        if Err<>ErrNoError then begin SaveTransparency:=Err;Exit;end;
      end;
    Data:=0;
    Err:=WriteChunk(Data,1);
    if Err<>ErrNoError then begin SaveTransparency:=Err;Exit;end;
  end
  else
  {savebpp is 24 or 32}
  begin
    {convert from imagebpp to 24/32 with _swap_}
    ConvertData(FarPtr(@PNGImage^.TransCol),FarPtr(@TransCol),PNGImage^.Pixel);
    {save red info (2 bytes - 1 is important)}
    i:=TransCol and $FF;
    SwapEndian16(i);
    Err:=WriteChunk(i,2);
    if Err<>ErrNoError then begin SaveTransparency:=Err;Exit;end;
    {save green info (2 bytes - 1 is important)}
    i:=TransCol shr 8 and $FF;
    SwapEndian16(i);
    Err:=WriteChunk(i,2);
    if Err<>ErrNoError then begin SaveTransparency:=Err;Exit;end;
    {save blue info (2 bytes - 1 is important)}
    i:=TransCol shr 16;
    SwapEndian16(i);
    Err:=WriteChunk(i,2);
    if Err<>ErrNoError then begin SaveTransparency:=Err;Exit;end;
  end;
  Err:=EndChunk;
  if Err<>ErrNoError then begin SaveTransparency:=Err;Exit;end;
end;

function SaveImageData : Word;
var
  Err : Word;
  Compress : Z_Stream;
  i : Word;
  MaxPackSize : Word;
  Inited : Boolean;

procedure PrepareExit(Code : Word);
begin
  if DataRow<>nil then
    FreeMem(DataRow,SaveLineSize);
  if LastRow<>nil then
    FreeMem(LastRow,SaveLineSize);
  if Row1<>nil then
    FreeMem(Row1,SaveLineSize+1);
  if Row2<>nil then
    FreeMem(Row2,MaxPackSize);
  if Inited then
  begin
    DeflateEnd(Compress);
    Inited:=False;
  end;
  SaveImageData:=Code;
end;

begin
  Inited:=False;
  Err:=StartChunk('IDAT');
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;

  MaxPackSize:=Round((SaveLineSize+1)*1.13);
  {alloc data buffers}
  GetMem(LastRow,SaveLineSize);
  if LastRow=nil then begin PrepareExit(ErrNoMem);Exit;end;
  GetMem(DataRow,SaveLineSize);
  if DataRow=nil then begin PrepareExit(ErrNoMem);Exit;end;
  GetMem(Row1,SaveLineSize+1);
  if Row1=nil then begin PrepareExit(ErrNoMem);Exit;end;
  GetMem(Row2,MaxPackSize);
  if Row2=nil then begin PrepareExit(ErrNoMem);Exit;end;

  {init compression}
  Fillbyte(farptr(LastRow),SaveLineSize,0);
  ImgPos:=PNGImage^.ImageStart;
  Compress.ZAlloc:=nil;
  Compress.ZFree:=nil;
  Compress.Opaque:=nil;
  Err:=DeflateInit(Compress,Z_Default_Compression);
  if Err<>Z_Ok then begin PrepareExit(ErrNoMem);Exit;end;
  Inited:=True;

  {compress and filter image data}
  for ScanLines:=1 to PNGImage^.SizeY do
  begin
    ConvertData(Ptr48(PNGImage^.Data,ImgPos),FarPtr(DataRow),LineSize);
    Row1^[0]:=4;
    ApplyEncodeFilter;
    Compress.Next_In:=@Row1^;
    Compress.Avail_In:=SaveLineSize+1;
    repeat
      Compress.Next_Out:=@Row2^;
      Compress.Avail_Out:=MaxPackSize;
      Err:=Deflate(Compress,Z_No_Flush);
      if Err<>Z_Ok then begin PrepareExit(Err);Exit;end;
      Err:=WriteChunk(Row2^,MaxPackSize-Compress.Avail_Out);
      if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
    until Compress.Avail_In=0;
    SwapLong(DataRow,LastRow);
    ImgPos:=ImgPos+PNGImage^.BPL;
  end;
  repeat
    Compress.Next_Out:=@Row2^;
    Compress.Avail_Out:=MaxPackSize;
    Err:=Deflate(Compress,Z_Finish);
    ScanLines:=WriteChunk(Row2^,MaxPackSize-Compress.Avail_Out);
    if ScanLines<>ErrNoError then begin PrepareExit(ScanLines);Exit;end;
  until Err<>Z_Ok;
  PrepareExit(ErrNoError);
  Err:=EndChunk;
  if Err<>ErrNoError then
  begin
    SaveImageData:=Err;
    Exit;
  end;
end;

function SavePNG(var Stream : TStream32;var Image : TImage) : Word;
var
  Err : Word;
  oldheaperrorfunc : pointer;

procedure PrepareExit(Code : Word);
begin
  SavePNG:=Code;
  HeapError:=OldHeapErrorFunc;
  PNGStream^.Done;
end;

begin
  savepng := errfileerror;
  if (stream.status <> stok) then exit;
  PNGImage:=@Image;
  PNGStream:=@Stream;
  oldheaperrorfunc := HeapError;
  HeapError := @MyHeapErrorFunc;
  SavePNG:=ErrNoError;
  with PNGHead do
  begin
    sizex:=PNGImage^.SizeX;
    SwapEndian32(sizex);
    sizey:=PNGImage^.SizeY;
    SwapEndian32(sizey);
    BitsPerSample:=PNGImage^.Bpp;
    if BitsPerSample>8 then
      BitsPerSample:=8;
    ColorType:=0;
    if PNGImage^.Flags and imGreyScale=0 then
      ColorType:=ColorType or ctColor;
    if PNGImage^.Flags and imAlpha<>0 then
      ColorType:=ColorType or ctAlpha;
    if PNGImage^.Flags and imPalette<>0 then
      ColorType:=ColorType or ctPalette;
    CM:=0;
    Filter:=0;
    IL:=0;
  end;
  SaveBpp:=PNGImage^.Bpp;
  if (SaveBpp in [15,16])or((SaveBpp=32)and(PNGHead.ColorType and ctAlpha=0)) then
    SaveBpp:=24;
  if PNGImage^.Bpp<>15 then
    LineSize:=PNGImage^.Bpp*PNGImage^.SizeX
  else
    LineSize:=16*PNGImage^.SizeX;
  if LineSize mod 8<>0 then
    Inc(LineSize,8);
  LineSize:=LineSize div 8;
  SaveLineSize:=SaveBpp*PNGImage^.SizeX;
  if SaveLineSize mod 8<>0 then
    Inc(SaveLineSize,8);
  SaveLineSize:=SaveLineSize div 8;
  Bypp:=SaveLineSize div PNGImage^.SizeX;
  BytesPerLine:=SaveLineSize;
  CNVTmp:=GetCNVProc(PNGImage^.Bpp,SaveBpp,cnvimgmd or cnvswap);
  ConvertData:=TCNVProc(CNVTmp);

  {write png header info}
  PNGStream^.Write32(FarPtr(@idpng),SizeOf(idpng));
  if PNGStream^.Status<>stOk then
  begin
    PrepareExit(StreamToGX2Err(PNGStream^));
    Exit;
  end;
  Err:=StartChunk('IHDR');
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
  Err:=WriteChunk(PNGHead,SizeOf(PNGHead));
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
  Err:=EndChunk;
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;

  {save data such as palette, transparency and image data}
  if PNGHead.ColorType and ctPalette<>0 then
  begin
    Err:=SavePalette;
    if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
  end;
  if (PNGImage^.Flags and imSprite<>0)and(PNGImage^.Flags and imAlpha=0) then
  begin
    Err:=SaveTransparency;
    if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
  end;
  Err:=SaveImageData;
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;

  {mark end of png file}
  Err:=StartChunk('IEND');
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;
  Err:=EndChunk;
  if Err<>ErrNoError then begin PrepareExit(Err);Exit;end;

  PrepareExit(errNoError);
end;

begin
end.
