{$A+,B-,F+,G+,N+,P+,Q-,R-,T-,V+,X+}
{****************************************************************************
 * Unit for GX2 to support printing                                         *
 * Not optimized, supports only few printers yet.                           *
 * Not free for commercial use without special license, for other copyright *
 * issues please refer grafx2.txt                                           *
 * Version 1.0                                                              *
 * .first unofficial version                                                *
 * Version 1.1                                                              *
 * .fixed several bugs in HP and ESC/P2 mode concerning missing and         *
 *  misplaced pixels in x-direction                                         *
 *                                                                          *
 ***** What, if you get the full license? ***********************************
 * If you get the full license for this unit, you'll get                    *
 * -a help file                                                             *
 * -better and faster error-diffusion                                       *
 * -full support for the unit                                               *
 * -probably more support for higher resolutions on Epson (depends on       *
 *  development state)                                                      *
 *                                                                          *
 ***** Hints ****************************************************************
 * -If the printer supports ESC/P2, use it - even if there seems to be no   *
 *  difference to ESC/P. Printing in ESC/P2 is faster                       *
 * -color error diffusion tends to look too dark and green - prefer other   *
 *  printing modes                                                          *
 * -If you want to use the full printer resolution, do the output to the    *
 *  printer in pieces of a height divideable by 48 - this requires less     *
 *  memory but should still be fast enough. Due to the full clipping in GX2,*
 *  this should be pretty easy                                              *
 ****************************************************************************}
unit gx_print;

interface

uses grafx2,newfrontier,gx_col,gx_glob,gx_image;

const
  lpt1 = $01;
  lpt2 = $02;
  lpt3 = $03;
  com1 = $10;
  com2 = $20;
  com3 = $30;
  com4 = $40;

  mdlhp300      = $0011;  { Deskjet Printer, 75..300 DPI       }
  mdlhp600      = $0012;  { Laser Printer, 75..600 DPI         }
  mdlhp1200     = $0013;  { Laser Printer, 75..1200 DPI        }

  mdlepson_esc  = $0021;  { Epson ESC/P Printer, 60..360 DPI   }
  mdlepson_esc2 = $0031;  { Epson ESC/P2 Printer, 180..360 DPI }

  mdlclbw       = $0000;
  mdlclcmy      = $0100;  { Color printer with CMY        }
  mdlclcmyk     = $0200;  { Color printer with CMYK       }

  paper_a4        = $00;
  paper_a5        = $01;
  paper_a6        = $02;
  paper_b5        = $03;
  paper_jis       = $03;
  paper_c5        = $04;
  paper_dl        = $05;
  paper_letter    = $06;
  paper_legal13   = $07;
  paper_legal14   = $08;
  paper_executive = $09;
  paper_com10     = $0A;
  paper_monarch   = $0B;

  { Printimage flags }
  pr_dither4x4   = $00;
  pr_dithererror = $01;

type
  tpaperspace = array[0..3] of double;

  pprinter = ^tprinter;
  tprinter = object
    constructor init(port,model : longint);
    destructor  done;

    function    initgraphic(resx,resy : longint) : boolean;
    procedure   donegraphic;

    procedure   setpapersize(paper : byte);
    function    printimage(var source : timage;scalex,scaley : double;brightness : integer;flags : longint) : boolean;
    function    finishpage : boolean;

    function    setprintpos(x,y : longint) : boolean;

    procedure   returnresolutiontable;
    function    getprintwidth : longint;
    function    getprintheight : longint;

    procedure   sendstring(s : string);

    private
    output   : file;
    vscreen  : tgxmem;
    ypos     : longint;    { Y position in source image data }
    colcount : byte;
    printermodel : longint;
    dpi          : tpoint;

    rdata     : tselector; { Raster data                     }
    scaledata : tselector;
    lastdata  : tselector; { Last scaled line                }
    linewidth : longint;   { Raster line width in bytes      }
    blocksize : tpoint;
    image     : pimage;
    bright    : integer;

    { Paper variables }
    currentpaper : byte;
    currentpapersize :
    record
      x,y : double;
    end;
    paperspace : tpaperspace;

    { Scale variables }
    umry,wx,old,sx,sy : longint;
    drawy             : longint; { Y position in output image data }
    imgptr            : pointer48;
    scale             : procedure(source,dest : pointer48;count,wx : longint);

    procedure   rasterblock_4x4_bw;
    procedure   rasterblock_4x4_color;
    procedure   rasterblock_error_bw;
    procedure   rasterblock_error_color;
    procedure   printblock;
  end;

implementation

const
  hpdonegfx = #27'*rB';
  hpreset   = #27'E';

  epsoninitgfx = #27'A'#8;
  epsondonegfx = #27'A'#10;
  epsonreset   = #27'@';

  epson2donegfx = #27'@';
  epson2reset   = #27'@';

  raster : array[0..7] of array[0..15] of byte =
  ((1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1),
   (0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1),
   (0,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1),
   (0,1,0,1,1,0,1,1,0,1,0,1,1,1,1,0),
   (0,1,0,1,1,0,1,0,0,1,0,1,1,0,1,0),
   (0,0,0,1,1,0,1,0,0,1,0,0,1,0,1,0),
   (0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0),
   (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0));

  mdlhp         = $10;
  mdlepson      = $20;
  mdlepson2     = $30;

  { Paper size array - values in inch }
  papersize : array[0..11,0..1] of double=(
  (8.27,11.69), { a4        }
  (5.83,8.27),  { a5        }
  (4.13,5.83),  { a6        }
  (7.17,10.12), { b5        }
  (6.38,9.02),  { c5        }
  (4.33,8.66),  { dl        }
  (8.5,11),     { letter    }
  (8.5,13),     { legal 13  }
  (8.5,14),     { legal 14  }
  (7.25,10.5),  { ececutive }
  (4.12,9.5),   { com-10    }
  (3.87,7.5)    { monarch   }
  );

  space_left   = 0;
  space_right  = 1;
  space_top    = 2;
  space_bottom = 3;

  space_hp     : tpaperspace=(0.25,0.25,0.18,0.12);
  { Hope these values affect all Epsons }
  space_epson  : tpaperspace=(0.2,0.2,0.34,0.54);
  space_epson2 : tpaperspace=(0.2,0.2,0.34,0.54);

  res_hp : array[0..3] of tpoint=(
  (x:75;y:75),
  (x:100;y:100),
  (x:150;y:150),
  (x:300;y:300)
  );

  res_epson : array[0..2] of tpoint=(
  (x:60;y:60),
  (x:180;y:180),
  (x:360;y:360)
  );

  res_epson2 : array[0..0] of tpoint=(
  (x:360;y:360)
  );

procedure binarymode(var dest);assembler;
asm
  les   di,dest
  MOV   BX,es:[di].word
  MOV   AX,4400H
  INT   21H
  OR    DL,20H
  MOV   DH,0
  MOV   AX,4401H
  INT   21H
end;

constructor tprinter.init(port,model : longint);
begin
  printermodel := model;
  assign(output,'LPT'+inttostr(port));
  rewrite(output,1);
  binarymode(output);
  case printermodel and $F0 of
    mdlhp     : begin
                  sendstring(hpreset);
                  paperspace := space_hp;
                end;
    mdlepson  : begin
                  sendstring(epsonreset);
                  paperspace := space_epson;
                end;
    mdlepson2 : begin
                  sendstring(epson2reset);
                  paperspace := space_epson2;
                end;
  end;
  setpapersize(paper_a4);

  case printermodel and $F00 of
    mdlclbw   : colcount := 1;
    mdlclcmy  : colcount := 3;
    mdlclcmyk : colcount := 4;
  end;
end;

destructor tprinter.done;
begin
  close(output);
end;

function tprinter.initgraphic(resx,resy : longint) : boolean;
begin
  dpi.x := resx;
  dpi.y := resy;
  case printermodel and $F0 of
    mdlhp     : begin
                  blocksize.y := 1;
                  sendstring(#27'*t'+inttostr(resx)+'R'#27'*r1A');
                  if colcount > 1 then sendstring(#27'*r-'+inttostr(colcount)+'U');
                {  sendstring(#27'&l0E');}
                  sendstring(#27'*b0M');
                end;
    mdlepson  : begin
                  blocksize.y := round(resy/7.5); { = used needles }
{                  case resy of
                    60  : blocksize.y := 8;
                    180 : blocksize.y := 24;
                    360 : blocksize.y := 48;
                  end;}
                  sendstring(epsoninitgfx);
                end;
    mdlepson2 : begin
                  blocksize.y := 1;
                  sendstring(#27'(G'#1#0#1);
                  sendstring(#27'(U'#1#0+chr(3600 div dpi.y));
{sendstring(#$1B#$28#$43#$02#$00#$E4#$20#$1B#$28#$63#$04#$00#$4E#$00#$94#$1F#$1B#$55#$01#$1B#$72#00#$1B#$28#$65#02#00#00#03);}
                  case printermodel and $F of
                    1 : sendstring(#27'+'+chr(360 div dpi.y));
                    2 : sendstring(#27'(K'#02#00#00#01);
                  end;
                end;
  end;
end;

procedure tprinter.donegraphic;
begin
  case printermodel and $F0 of
    mdlhp     : sendstring(hpdonegfx);
    mdlepson  : sendstring(epsondonegfx);
    mdlepson2 : sendstring(epson2donegfx);
  end;
end;

procedure tprinter.setpapersize(paper : byte);
var
  b : byte;
begin
  currentpaper := paper;
  currentpapersize.x := papersize[paper,0];
  currentpapersize.y := papersize[paper,1];
{!!}
{  if (printermodel and $F0 = mdlhp) then begin
    case paper of
      paper_a4        : b := 26;
      paper_a5        : b := 27;
      paper_a6        : b := 28;
      paper_b5        : b := 37;
      paper_c5        : b := 91;
      paper_dl        : b := 90;
      paper_letter    : b := 2;
      paper_legal13   : b := 4;
      paper_legal14   : b := 3;
      paper_executive : b := 1;
      paper_com10     : b := 81;
      paper_monarch   : b := 80;
    end;
    sendstring(#27'&1'+chr(b)+'A');
  end;}
end;

function tprinter.printimage(var source : timage;scalex,scaley : double;brightness : integer;flags : longint) : boolean;
var
  y : longint;
begin
  image := @source;
  printimage := false;
  if (scalex*scaley = 0) then exit;
  sx     := round(image^.sizex*scalex);
  sy     := round(image^.sizey*scaley);
  wx     := (65536*image^.sizex-1) div sx;
  umry   := 0;
  drawy  := 0;
  ypos   := 0;
  old    := $FFFF;
  imgptr := ptr48(image^.data,image^.imagestart);
  bright := brightness;

  case image^.bpp of
    8     : scale := scale_8;
    15,16 : scale := scale_1516;
    24    : scale := scale_24;
    32    : scale := scale_32;
    else exit;
  end;

  vscreen.init(image^);
  blocksize.x := sx;
  linewidth   := (blocksize.x+7) shr 3;
  getmem32(rdata,(linewidth*blocksize.y+1)*colcount);
  getmem32(scaledata,(sx+1)*image^.pixel+4);
  getmem32(lastdata,(sx+1)*image^.pixel+4);
  for y := 0 to sy div blocksize.y-1 do
  begin
    case flags and $01 of
      pr_dither4x4 : if (colcount = 1) then rasterblock_4x4_bw else rasterblock_4x4_color;
      pr_dithererror : if (colcount = 1) then rasterblock_error_bw else rasterblock_error_color;
    end;
    printblock;
    inc(drawy,blocksize.y);
  end;
  freemem32(lastdata);
  freemem32(scaledata);
  freemem32(rdata);
  vscreen.close;
  printimage := true;
end;

function tprinter.finishpage : boolean;
begin
  sendstring(#12);
  finishpage := true;
end;

function tprinter.setprintpos(x,y : longint) : boolean;
begin
end;

procedure tprinter.returnresolutiontable;
begin
end;

function tprinter.getprintheight : longint;
begin
  getprintheight := trunc((currentpapersize.y-paperspace[space_top]-
                           paperspace[space_bottom])*dpi.y);
end;

function tprinter.getprintwidth : longint;
begin
  getprintwidth := trunc((currentpapersize.x-paperspace[space_left]-
                          paperspace[space_right])*dpi.x);
end;

function get(sel : tselector;x,bpp : longint) : longint;
var
  rgb : trgb;
  col : longint;
begin
  case bpp of
    15,16 : col := readmem32w(ptr48(sel,x shl 1));
    24    : col := readmem32l(ptr48(sel,x*3)) and $FFFFFF;
    32    : col := readmem32l(ptr48(sel,x shl 2));
  end;
  case bpp of
    15 : dergb_15(col,rgb);
    16 : dergb_16(col,rgb);
    24,32 : dergb_2432(col,rgb);
    else rgb.l := $FFFFFFFF;
  end;
  get := rgb.l;
end;

procedure put(sel : tselector;x,bpp : longint;col : trgb);
begin
  case bpp of
    15 : writemem32w(ptr48(sel,x shl 1),rgb_15(col.r,col.g,col.b));
    16 : writemem32w(ptr48(sel,x shl 1),rgb_16(col.r,col.g,col.b));
    24 : begin
           writemem32w(ptr48(sel,x*3),rgb_2432(col.r,col.g,col.b));
           writemem32b(ptr48(sel,x*3+2),rgb_2432(col.r,col.g,col.b) shr 16);
         end;
    32 : writemem32l(ptr48(sel,x shl 2),rgb_2432(col.r,col.g,col.b));
  end;
end;

procedure tprinter.rasterblock_4x4_bw;
const
  vr = 35;
  vg = 40;
  vb = 25;

var
  x,y     : longint;
  val,col : integer;
  rgb     : trgb;
  pos,r   : byte;
  mempos  : longint;

begin
  fillchar(ptr(rdata,0)^,linewidth*blocksize.y,0);
  for y := 0 to blocksize.y-1 do begin
    val := 0;
    if (y+drawy > sy) then break;

    if (umry div sy <> old) then begin
      scale(imgptr,ptr48(scaledata,0),sx*image^.pixel,wx);
      if (y+drawy > 0) then
       inc(tselofs32(imgptr).offset32,image^.bpl*(umry div sy-old))
       else inc(tselofs32(imgptr).offset32,image^.bpl);
      old := umry div sy;
    end;
    inc(umry,image^.sizey);

    for x := 0 to blocksize.x-1 do
    begin
      rgb.l := get(scaledata,x,image^.bpp);
      col := (rgb.r*vr+rgb.g*vg+rgb.b*vb) div 100+15+bright;
      if (col > 255) then col := 255;
      if (col < 0) then col := 0;
      pos := (x and 3)+((y+drawy) and 3) shl 2;
      r   := raster[col shr 5,pos];
      case printermodel and $F0 of
        mdlhp,mdlepson2 :
        begin
          if (x and 7 = 0)and(x > 0) then
          begin
            writemem32b(ptr48(rdata,(x-1) shr 3+y*linewidth),val);
            val := 0;
          end;
          val := val or (r shl (7-x and 7));
        end;
        mdlepson :
        begin
          mempos := (y shr 3)+x*(blocksize.y shr 3);
          mem[rdata:mempos] := mem[rdata:mempos] or (r shl (7-y and 7));
        end;
      end;
    end;
    if (printermodel and $F0 = mdlhp)or(printermodel and $F0 = mdlepson2)and(x and 3 <> 0) then
      writemem32b(ptr48(rdata,x shr 3+y*linewidth),val);
  end;
end;

procedure tprinter.rasterblock_4x4_color;
var
  x,y : longint;
  r,i : integer;
  rgb : trgb;
  pos : byte;
  col : array[0..3] of byte;
  mempos : longint;
  val : array[0..3] of byte;

begin
  fillchar(ptr(rdata,0)^,linewidth*blocksize.y*colcount,0);
  for y := 0 to blocksize.y-1 do begin
    for i := 0 to 3 do val[i] := 0;
    if (y+drawy > sy) then break;

    if (umry div sy <> old) then begin
      scale(imgptr,ptr48(scaledata,0),sx*image^.pixel,wx);
      if (y+drawy > 0) then
       inc(tselofs32(imgptr).offset32,image^.bpl*(umry div sy-old))
       else inc(tselofs32(imgptr).offset32,image^.bpl);
      old := umry div sy;
    end;
    inc(umry,image^.sizey);

    for x := 0 to blocksize.x-1 do
    begin
      rgb.l := get(scaledata,x,image^.bpp);

      r := rgb.r+bright;
      if (r > 255) then r := 255;if (r < 0) then r := 0;
      col[0] := r;
      r := rgb.g+bright;
      if (r > 255) then r := 255;if (r < 0) then r := 0;
      col[1] := r;
      r := rgb.b+bright;
      if (r > 255) then r := 255;if (r < 0) then r := 0;
      col[2] := r;

      pos := (x and 3)+((y+drawy) and 3) shl 2;
      for i := 0 to 2 do col[i] := raster[col[i] shr 5,pos];

      if (colcount >= 4) then begin
        col[3] := col[0] and col[1] and col[2];
        if (col[3] = 1) then for i := 0 to 2 do col[i] := 0;
      end;

      case printermodel and $F0 of
        mdlhp,mdlepson2 :
        begin
          if (x and 7 = 0)and(x > 0) then
          for i := 0 to colcount-1 do begin
            writemem32b(ptr48(rdata,(x-1) shr 3+y*linewidth+linewidth*blocksize.y*i),val[i]);
            val[i] := 0;
          end;
          for i := 0 to colcount-1 do val[i] := val[i] or (col[i] shl (7-x and 7));
        end;
        mdlepson :
        for i := 0 to colcount-1 do
        begin
          if col[i] = 1 then begin
            mempos := (y shr 3)+x*(blocksize.y shr 3)+linewidth*blocksize.y*i;
            mem[rdata:mempos] := mem[rdata:mempos] or (col[i] shl (7-y and 7));
          end;
        end;
      end;
    end;
    if (printermodel and $F0 = mdlhp)or(printermodel and $F0 = mdlepson2)and(x and 3 <> 0) then
     begin
       for i := 0 to colcount-1 do val[i] := val[i] or (col[i] shl (7-x and 7));
       for i := 0 to colcount-1 do begin
         writemem32b(ptr48(rdata,x shr 3+y*linewidth+linewidth*blocksize.y*i),val[i]);
         val[i] := 0;
       end;
     end;
  end;
end;


procedure pmod(sel : tselector;x,bpp,index,rgb : longint);
var
  col : trgb;
  l   : longint;
begin
  case bpp of
    15,16 : l := readmem32w(ptr48(sel,x shl 1));
    24    : l := readmem32l(ptr48(sel,x*3)) and $FFFFFF;
    32    : l := readmem32l(ptr48(sel,x shl 2));
  end;
  case bpp of
    15 : dergb_15(l,col);
    16 : dergb_16(l,col);
    24,32 : dergb_2432(l,col);
  end;

  col.col[index] := rgb;

  case bpp of
    15 : writemem32w(ptr48(sel,x shl 1),rgb_15(col.r,col.g,col.b));
    16 : writemem32w(ptr48(sel,x shl 1),rgb_16(col.r,col.g,col.b));
    24 : begin
           writemem32w(ptr48(sel,x*3),rgb_2432(col.r,col.g,col.b));
           writemem32b(ptr48(sel,x*3+2),rgb_2432(col.r,col.g,col.b) shr 16);
         end;
    32 : writemem32l(ptr48(sel,x shl 2),rgb_2432(col.r,col.g,col.b));
  end;
end;

function getcl(sel : tselector;x,bpp,index : longint) : integer;
var
  rgb : trgb;
  col : longint;
begin
  case bpp of
    15,16 : col := readmem32w(ptr48(sel,x shl 1));
    24    : col := readmem32l(ptr48(sel,x*3)) and $FFFFFF;
    32    : col := readmem32l(ptr48(sel,x shl 2));
  end;
  case bpp of
    15 : dergb_15(col,rgb);
    16 : dergb_16(col,rgb);
    24,32 : dergb_2432(col,rgb);
    else rgb.l := $FFFFFFFF;
  end;
  getcl := rgb.col[index];
end;


procedure tprinter.rasterblock_error_bw;
const
  vr = 35;
  vg = 40;
  vb = 25;

var
  x,y      : longint;
  r        : integer;
  rgb,rgb2 : trgb;
  pos      : byte;
  col      : byte;
  mempos   : longint;
  val      : byte;
  k,error  : integer;

begin
  fillchar(ptr(rdata,0)^,linewidth*blocksize.y*colcount,0);
  for y := 0 to blocksize.y-1 do begin
    if (y+drawy > sy) then break;
    val := 0;
    move2(ptr48(scaledata,0),ptr48(lastdata,0),(sx+1)*image^.pixel+4);
    scale(imgptr,ptr48(scaledata,0),sx*image^.pixel,wx);
    for x := 0 to sx do
    begin
      rgb.l := get(scaledata,x,image^.bpp);
      col := (rgb.r*vr+rgb.g*vg+rgb.b*vb) div 400;
      rgb.l := col;
      put(scaledata,x,image^.bpp,rgb);
    end;
    if y+drawy = 0 then move2(ptr48(scaledata,0),ptr48(lastdata,0),(sx+1)*image^.pixel+4);
    if (umry div sy <> old) then begin
      if (y+drawy > 0) then
       inc(tselofs32(imgptr).offset32,image^.bpl*(umry div sy-old))
       else inc(tselofs32(imgptr).offset32,image^.bpl);
      old := umry div sy;
    end;
    inc(umry,image^.sizey);

    for x := 1 to blocksize.x-2 do
    begin
      rgb.l := get(lastdata,x,image^.bpp);

      k := rgb.l+bright div 4;

      if k > 31 then k := 63 else k := 0;

      if k > 0 then col := 0 else col := 1;

      error := rgb.col[0]-k;
      rgb.l := getcl(lastdata,x+1,image^.bpp,0)+7 * error div 16;
      if rgb.l < 0 then rgb.l := 0;
      put(lastdata,x+1,image^.bpp,rgb);
      rgb.l := getcl(scaledata,x-1,image^.bpp,0)+3 * error div 16;
      if rgb.l < 0 then rgb.l := 0;
      put(scaledata,x-1,image^.bpp,rgb);
      rgb.l := getcl(scaledata,x,image^.bpp,0)+5 * error div 16;
      if rgb.l < 0 then rgb.l := 0;
      put(scaledata,x,image^.bpp,rgb);
      rgb.l := getcl(scaledata,x+1,image^.bpp,0)+error-(15 * error div 16);
      if rgb.l < 0 then rgb.l := 0;
      put(scaledata,x+1,image^.bpp,rgb);

      case printermodel and $F0 of
        mdlhp,mdlepson2 :
        begin
          if (x and 7 = 0)and(x > 0) then
          begin
            writemem32b(ptr48(rdata,(x-1) shr 3+y*linewidth),val);
            val := 0;
          end;
          val := val or (col shl (7-x and 7));
        end;
        mdlepson :
        begin
          if col = 1 then begin
            mempos := (y shr 3)+x*(blocksize.y shr 3);
            mem[rdata:mempos] := mem[rdata:mempos] or (col shl (7-y and 7));
          end;
        end;
      end;
    end;
    if (printermodel and $F0 = mdlhp)or(printermodel and $F0 = mdlepson2)and(x and 3 <> 0) then
     begin
       val := val or (col shl (7-x and 7));
       if (x and 7 = 0)and(x > 0) then
       begin
         writemem32b(ptr48(rdata,x shr 3+y*linewidth+linewidth*blocksize.y),val);
         val := 0;
       end;
     end;
  end;
end;

procedure tprinter.rasterblock_error_color;
var
  x,y      : longint;
  r,i,cl   : integer;
  rgb,rgb2 : trgb;
  pos      : byte;
  col      : array[0..3] of byte;
  mempos   : longint;
  val      : array[0..3] of byte;
  k,error  : integer;

begin
  fillchar(ptr(rdata,0)^,linewidth*blocksize.y*colcount,0);
  for y := 0 to blocksize.y-1 do begin
    if (y+drawy > sy) then break;
    for i := 0 to 3 do val[i] := 0;

    move2(ptr48(scaledata,0),ptr48(lastdata,0),(sx+1)*image^.pixel+4);
    scale(imgptr,ptr48(scaledata,0),sx*image^.pixel,wx);
    for x := 0 to sx do
    begin
      rgb.l := get(scaledata,x,image^.bpp);
      for i := 0 to 2 do rgb.col[i] := rgb.col[i] shr 2;
      put(scaledata,x,image^.bpp,rgb);
    end;
    if y+drawy = 0 then move2(ptr48(scaledata,0),ptr48(lastdata,0),(sx+1)*image^.pixel+4);
    if (umry div sy <> old) then begin
      if (y+drawy > 0) then
       inc(tselofs32(imgptr).offset32,image^.bpl*(umry div sy-old))
       else inc(tselofs32(imgptr).offset32,image^.bpl);
      old := umry div sy;
    end;
    inc(umry,image^.sizey);

    for x := 1 to blocksize.x-2 do
    begin

      rgb.l := get(lastdata,x,image^.bpp);
      for i := 0 to 2 do
      begin
        cl := rgb.col[i]+bright div 4;

        if cl > 31 then k := 63 else k := 0;

        if k > 0 then col[2-i] := 0 else col[2-i] := 1;

        error := rgb.col[i]-k;
        cl := getcl(lastdata,x+1,image^.bpp,i)+7 * error div 16;
        if cl < 0 then cl := 0;
        pmod(lastdata,x+1,image^.bpp,i,cl);
        cl := getcl(scaledata,x-1,image^.bpp,i)+3 * error div 16;
        if cl < 0 then cl := 0;
        pmod(scaledata,x-1,image^.bpp,i,cl);
        cl := getcl(scaledata,x,image^.bpp,i)+5 * error div 16;
        if cl < 0 then cl := 0;
        pmod(scaledata,x,image^.bpp,i,cl);
        cl := getcl(scaledata,x+1,image^.bpp,i)+error-(15 * error div 16);
        if cl < 0 then cl := 0;
        pmod(scaledata,x+1,image^.bpp,i,cl);
      end;

      if (colcount >= 4) then begin
        col[3] := col[0] and col[1] and col[2];
        if (col[3] = 1) then for i := 0 to 2 do col[i] := 0;
      end;

      case printermodel and $F0 of
        mdlhp,mdlepson2 :
        begin
          if (x and 7 = 0)and(x > 0) then
          for i := 0 to colcount-1 do begin
            writemem32b(ptr48(rdata,(x-1) shr 3+y*linewidth+linewidth*blocksize.y*i),val[i]);
            val[i] := 0;
          end;
          for i := 0 to colcount-1 do val[i] := val[i] or (col[i] shl (7-x and 7));
        end;
        mdlepson :
        for i := 0 to colcount-1 do
        begin
          if col[i] = 1 then begin
            mempos := (y shr 3)+x*(blocksize.y shr 3)+linewidth*blocksize.y*i;
            mem[rdata:mempos] := mem[rdata:mempos] or (col[i] shl (7-y and 7));
          end;
        end;
      end;
    end;
    if (printermodel and $F0 = mdlhp)or(printermodel and $F0 = mdlepson2)and(x and 3 <> 0) then
     begin
       for i := 0 to colcount-1 do val[i] := val[i] or (col[i] shl (7-x and 7));
       for i := 0 to colcount-1 do begin
         writemem32b(ptr48(rdata,x shr 3+y*linewidth+linewidth*blocksize.y*i),val[i]);
         val[i] := 0;
       end;
     end;
  end;
end;

procedure tprinter.printblock;
var
  i,i2,val : byte;

const
  epsoncol : array[0..3] of char = (#2,#1,#4,#0);
begin
  case printermodel and $F0 of
    mdlhp   : for i := 0 to blocksize.y-1 do begin
                if (colcount = 1) then
                begin
                  sendstring(#27'*b'+inttostr(linewidth)+'W');
                  blockwrite(output,ptr(rdata,i*linewidth)^,linewidth);
                end else
                begin
                  for i2 := 0 to colcount-1 do begin
                    if (i2 < colcount-1) then
                    sendstring(#27'*b'+inttostr(linewidth)+'V')
                     else sendstring(#27'*b'+inttostr(linewidth)+'W');
                     blockwrite(output,ptr(rdata,i*linewidth+linewidth*blocksize.y*i2)^,linewidth);
                  end;
                end;
              end;
    mdlepson: begin
                for i := 0 to colcount-1 do
                begin
                  case dpi.x of
                    60  : val := 0;
                    180 : val := 39;
                    360 : val := 73;
                  end;
                  if (colcount > 1) then sendstring(#27'r'+epsoncol[i]);
                  sendstring(#27'*'+chr(val)+chr(sx and $FF)+chr(sx shr 8));
                  blockwrite(output,ptr(rdata,linewidth*blocksize.y*i)^,linewidth*blocksize.y);
                  sendstring(#13);
                end;
                sendstring(#10);
              end;
    mdlepson2:begin
                for i := 0 to colcount-1 do
                begin
                  if (colcount > 1) then sendstring(#27'r'+epsoncol[i]);
                  sendstring(#27#46#0+chr(3600 div dpi.x)+chr(3600 div dpi.y)+#01+chr(sx and $FF)+chr(sx shr 8));
                  blockwrite(output,ptr(rdata,linewidth*blocksize.y*i)^,linewidth);
                  sendstring(#13);
                end;
                sendstring(#10);
              end;
  end;
end;

procedure tprinter.sendstring(s : string);
begin
  blockwrite(output,s[1],length(s));
end;

begin
end.